"""
Module to compute centroidal voronoi tesselation mesh from an existing mesh of
convex cells.

In principle, this is easy to compute:

* Compute the centroids of the cells.
* Invert the face_node_connectivity index array.
* For every node, find the connected faces.
* Use the connected faces to find the centroids.
* Order the centroids around the vertex in a counter-clockwise manner.

Dealing with the mesh exterior (beyond which no centroids are located) is the
tricky part. Refer to the docstrings and especially the visual examples in the
developer documentation.
"""

from typing import Tuple

import numpy as np
import pandas as pd
from scipy import sparse

from xugrid.constants import FILL_VALUE, X_EPSILON, FloatArray, IntArray
from xugrid.ugrid.connectivity import (
    area_from_coordinates,
    close_polygons,
    ragged_index,
    renumber,
)


def dot_product2d(U: FloatArray, V: FloatArray):
    return U[:, 1] * V[:, 1] + U[:, 0] * V[:, 0]


def _centroid_pandas(i: IntArray, x: FloatArray, y: FloatArray):
    grouped = pd.DataFrame({"i": i, "x": x, "y": y}).groupby("i").mean()
    x_centroid = grouped["x"].to_numpy()
    y_centroid = grouped["y"].to_numpy()
    return x_centroid, y_centroid


def _centroid_scipy(i: IntArray, x: FloatArray, y: FloatArray):
    j = np.arange(len(i))
    coo_content = (x, (i, j))
    mat = sparse.coo_matrix(coo_content)
    n = mat.getnnz(axis=1)
    x_centroid = mat.sum(axis=1).flat / n
    mat.data = y
    y_centroid = mat.sum(axis=1).flat / n
    return x_centroid, y_centroid


def compute_centroid(i: IntArray, x: FloatArray, y: FloatArray):
    return _centroid_pandas(i, x, y)


def _create_face_node_connectivity(i: IntArray, j: IntArray) -> IntArray:
    n_vertex = np.bincount(i)
    n_vertex = n_vertex[n_vertex > 0]
    n = len(n_vertex)
    m = n_vertex.max()
    index = ragged_index(n, m, n_vertex)
    face_node_connectivity = np.full((n, m), FILL_VALUE)
    face_node_connectivity[index] = j
    return face_node_connectivity


def exterior_centroids(node_face_connectivity: sparse.csr_matrix):
    n, _ = node_face_connectivity.shape
    # Find exterior nodes NOT associated with any interior edge
    is_exterior_only = node_face_connectivity.getnnz(axis=1) == 1
    j = node_face_connectivity[is_exterior_only].indices
    i = np.arange(n)[is_exterior_only]
    return i, j


def interior_centroids(
    node_face_connectivity: sparse.csr_matrix,
    edge_face_connectivity: IntArray,
    edge_node_connectivity: IntArray,
):
    # Find exterior nodes associated with interior edges
    is_exterior = edge_face_connectivity[:, 1] == FILL_VALUE
    exterior_nodes = np.unique(edge_node_connectivity[is_exterior].ravel())
    m_per_node = node_face_connectivity.getnnz(axis=1)
    is_interior_only = m_per_node > 1

    selected_nodes = exterior_nodes[is_interior_only[exterior_nodes]]
    selection = node_face_connectivity[selected_nodes]
    m_per_selected_node = selection.getnnz(axis=1)

    j = selection.indices
    i = np.repeat(selected_nodes, repeats=m_per_selected_node)
    return i, j


def _project_centroid_on_edge(
    edge_vertices: FloatArray, centroid_vertices: FloatArray
) -> FloatArray:
    a = edge_vertices[:, 0, :]
    b = edge_vertices[:, 1, :]
    V = b - a
    U = centroid_vertices - a
    # np.dot(U, V) doesn't do the desired thing here
    projected_vertices = a + ((dot_product2d(U, V) / dot_product2d(V, V)) * V.T).T
    return projected_vertices


def _filter_projections(
    projected_vertices: FloatArray,
    centroid_vertices: FloatArray,
    exterior_nodes: IntArray,
    face_i: IntArray,
) -> Tuple[IntArray, FloatArray, IntArray]:
    # Create the numbering pointing to these new vertices.
    # Discard vertices that overlap with e.g. circumcenters.
    keep = np.linalg.norm(projected_vertices - centroid_vertices, axis=1) > (
        X_EPSILON * X_EPSILON
    )
    face_i = face_i[keep]
    vertices = projected_vertices[keep]
    i = exterior_nodes[keep].ravel()
    return i, vertices, face_i


def _interpolate_between_projections(
    projected_vertices: FloatArray,
    exterior_nodes: IntArray,
    n: int,
) -> Tuple[IntArray, IntArray, FloatArray, IntArray]:
    n_new = len(projected_vertices)
    i = exterior_nodes.ravel()
    order = np.argsort(i)
    jj = np.repeat(np.arange(n_new), 2)[order]
    to_interpolate = projected_vertices[jj]
    interpolated_vertices = 0.5 * (to_interpolate[::2] + to_interpolate[1::2])
    # For the interpolated vertices, it depends on the two other nodes (which
    # depend on centroids). Store for each interpolation on which two nodes it relies.
    j = np.arange(n, n + len(interpolated_vertices))
    interpolation_map = jj.reshape((-1, 2))
    return i[order][::2], j, interpolated_vertices, interpolation_map


def exterior_vertices(
    edge_face_connectivity: IntArray,
    edge_node_connectivity: IntArray,
    vertices: FloatArray,
    centroids: FloatArray,
    add_vertices: bool,
):
    is_exterior = edge_face_connectivity[:, 1] == FILL_VALUE
    exterior_nodes = edge_node_connectivity[is_exterior]
    # For every exterior node, project the centroids to exterior edges
    edge_vertices = vertices[exterior_nodes]
    face_i = edge_face_connectivity[is_exterior, 0]
    centroid_vertices = centroids[face_i]
    projected_vertices = _project_centroid_on_edge(edge_vertices, centroid_vertices)
    i, vertices, face_i = _filter_projections(
        projected_vertices,
        centroid_vertices,
        exterior_nodes,
        face_i,
    )

    n_centroid = len(centroids)
    n = n_centroid + len(vertices)
    j = np.repeat(np.arange(n_centroid, n), 2)
    n_interpolated = 0
    interpolation_map = None

    # We add a guaranteed convex substitution vertex instead of the actual vertex
    if add_vertices:
        (
            i_new,
            j_new,
            interpolated_vertices,
            interpolation_map,
        ) = _interpolate_between_projections(projected_vertices, exterior_nodes, n)
        interpolation_map += n_centroid
        n_interpolated = len(interpolated_vertices)
        i = np.concatenate([i, i_new])
        j = np.concatenate([j, j_new])
        vertices = np.concatenate([vertices, interpolated_vertices])
        # Create face index: the face of the original mesh associated with every
        # voronoi vertex is a centroid. The exterior vertices are an exception to
        # this: these are associated with two faces. So we set a value of -1 here.
        face_i = np.concatenate([face_i, np.full(n_interpolated, -1)])

    return i, j, vertices, face_i, n_interpolated, interpolation_map


def choose_convex(
    i: IntArray,
    j: IntArray,
    nodes: FloatArray,
    original_vertices: FloatArray,
    n_interpolated: int,
) -> None:
    # Determine whether the original vertex or the interpolated vertex
    # generates the largest area (since the concave face will be smaller than
    # the convex face.)
    # Create a face_node_connectivity array.
    n_vertex = np.bincount(i)
    n_vertex = n_vertex[n_vertex > 0]
    n = len(n_vertex)
    m = n_vertex.max()
    index = ragged_index(n, m, n_vertex)
    faces = np.full((n, m), -1)
    faces[index] = j
    # Close the polygons so we can easily compute areas.
    closed, _ = close_polygons(faces)
    # Make a copy and insert the original vertices.
    modified_nodes = nodes.copy()
    modified_nodes[-n_interpolated:] = original_vertices

    # Compare areas of faces
    convex_area = area_from_coordinates(nodes[closed])
    modified_area = area_from_coordinates(modified_nodes[closed])
    original_is_convex = (modified_area >= convex_area)[:, np.newaxis]
    # All the newly created vertices are found at the end.
    is_interpolated = faces >= len(nodes) - n_interpolated
    # No need for unique: every exterior vertex is featured exactly once.
    use_original = faces[original_is_convex & is_interpolated]
    nodes[use_original] = modified_nodes[use_original]
    return


def exterior_topology(
    edge_face_connectivity: IntArray,
    edge_node_connectivity: IntArray,
    node_face_connectivity: sparse.csr_matrix,
    vertices: FloatArray,
    centroids: FloatArray,
    add_vertices: bool,
    skip_concave: bool,
):
    """
    Create the exterior topology of the voronoi tesselation.

    The exterior topology of this voronoi tesselation consists of three kinds
    of vertices:

    * Centroids of faces that have a vertex located on the exterior, but have
      no exterior edges. These centroids are used twice in the resulting
      voronoi mesh topology.
    * Centroids of faces that have an exterior edge. These centroids are used
      once.
    * Vertices of the intersection of (infinite) rays with the mesh exterior.
      These vertices are used twice; these intersections are the orthogonal
      projection of the centroid of the face on its exterior edges.
    * Original vertices of the mesh exterior. These vertices are used once, and
      are always located in between the projected vertices.

    The last ones are trickiest, as they may create concave angles; with a
    concave angle in a polygon, we cannot easily compute the counter clockwise
    order of the vertices to form a polygon. However, since these vertices are
    always located in between the projected vertices, we can introduce a
    subsitute: a linear interpolation of the two projected vertices. We
    can then utilize this interpolated vertex to compute the counter clockwise
    order, and then replace it by the original vertex -- which then may
    introduce a concave angle.

    A note on the index arrays:

    * i refers to the node number of the original mesh.
    * In principle, j refers to the face number (and associated centroids) of
    the original mesh; exterior vertices require new numbers as they are not a
    centroid.

    Finally, in the resulting new (voronoi) face_node_connectivity, i becomes
    the face number, and j becomes the node number.
    """
    i0, j0 = interior_centroids(
        node_face_connectivity,
        edge_face_connectivity,
        edge_node_connectivity,
    )
    i1, j1 = exterior_centroids(node_face_connectivity)
    (
        i2,
        j2,
        projected_vertices,
        face_i,
        n_interpolated,
        interpolation_map,
    ) = exterior_vertices(
        edge_face_connectivity,
        edge_node_connectivity,
        vertices,
        centroids,
        add_vertices,
    )

    i = np.concatenate([i0, i1, i2])
    j = np.concatenate([j0, j1, j2])
    _, n_face = node_face_connectivity.shape
    vor_vertices = np.concatenate([centroids, projected_vertices])
    face_i = np.concatenate([np.arange(n_face), face_i])
    orig_vertices = vertices[i][-n_interpolated:]

    # Now form valid counter-clockwise polygons
    voronoi_vertices = vor_vertices[j]
    x = voronoi_vertices[:, 0]
    y = voronoi_vertices[:, 1]
    centroid_x, centroid_y = compute_centroid(i, x, y)
    renumbered_i = renumber(i)
    dx = x - centroid_x[renumbered_i]
    dy = y - centroid_y[renumbered_i]
    angle = np.arctan2(dy, dx)
    # Sort to create a counter clockwise oriented polygon.
    # Use a lexsort to make sure nodes are grouped together.
    order = np.lexsort((angle, i))
    i = i[order]
    j = j[order]

    # If add_vertices is True, we have substituted interpolated points before
    # to generate the proper ordering. We overwrite those substituted points
    # here by their possibly concave true vertices.
    if add_vertices:
        if skip_concave:
            choose_convex(i, j, vor_vertices, orig_vertices, n_interpolated)
        else:
            vor_vertices[-n_interpolated:] = orig_vertices

    return vor_vertices, i, j, face_i, interpolation_map


def voronoi_topology(
    node_face_connectivity: sparse.csr_matrix,
    vertices: FloatArray,
    centroids: FloatArray,
    edge_face_connectivity: IntArray = None,
    edge_node_connectivity: IntArray = None,
    add_exterior: bool = False,
    add_vertices: bool = False,
    skip_concave: bool = False,
) -> Tuple[FloatArray, sparse.csr_matrix, IntArray, IntArray]:
    """
    Compute the centroidal voronoi tesslation (CVT) of an existing mesh of
    (convex!) cells using connectivity index arrays.

    If the exterior boundary of the mesh forms a concave polygon, this will
    result in some concave voronoi cells as well. Since concave cells are often
    undesirable, there are three options for dealing with the exterior
    boundary:

    * if ``add_exterior=True`` and ``add_vertices=True``, infinite voronoi rays
      (projections) are intersected at exterior edges and all exterior vertices
      are included. If the mesh exterior forms a concave polygon, the resulting
      tesselation will contain concave cells as well.
    * if ``add_exterior=True`` and ``add_vertices=False``, the infinite voronoi
      rays are intersected, but no exterior vertices are included. This will
      always result in convex voronoi cells if the original cells are convex.
    * if ``add_exterior=False`` and ``add_vertices=False``, no vertices on the
      exterior edges are considered. Only centroids of the original mesh are
      considered. This will always result in convex voronoi cells if the
      original cells are convex.

    A direct correspondence exists between the voronoi vertices and the
    original faces: the vertices of the voronoi polygons are the centroids of
    the original faces; and the intersected rays "belong" to a single face too.
    This is not the case when all exterior vertices are included: these may
    correspond to multiple faces of the original mesh. These vertices are
    marked by a face_index of -1.

    Parameters
    ----------
    node_face_connectivity : csr matrix
    vertices: ndarray of floats with shape ``(n_vertex, 2)``
    centroids: ndarray of floats with shape ``(n_centroid, 2)``
    edge_face_connectivity: ndarray of integers with shape ``(n_edge, 2)``, optional
    edge_node_connectivity: ndarray of integers with shape ``(n_edge, 2)``, optional
    add_exterior: bool, optional
        Whether to consider exterior edges of the original mesh, or to consider
        exclusively centroids.
    add_vertices: bool, optional
        Whether to use existing exterior vertices.
    skip_concave: bool, optional
        Whether to skip existing exterior vertices if they generate concave
        faces.

    Returns
    -------
    nodes: ndarray of floats with shape ``(n_vertex, 2)``
    face_node_connectivity: scipy.sparse.csr_matrix
    face_index: ndarray of ints with shape ``(n_vertex,)``
        Connects the nodes of the voronoi topology to the faces of the original
        grid. Exterior vertices (when ``add_vertices=True``) are given an index
        of -1.
    interpolation_map: ndarray of ints with shape ``(n_interpolated, 2)``
        Marks for each interpolated point from which nodes it has been
        interpolated.
    """
    if add_exterior:
        if any(arg is None for arg in [edge_face_connectivity, edge_node_connectivity]):
            raise ValueError(
                "edge_face_connectivity, edge_node_connectivity "
                "must be provided if add_exterior is True."
            )

    # Avoid overlapping polygons: if exterior is included, the exterior
    # algorithm will construct those polygons. If exterior is not included, we
    # take any valid internal polygon we can construct: at least a triangle.
    ncol_per_row = node_face_connectivity.getnnz(axis=1)
    if add_exterior:
        is_exterior = edge_face_connectivity[:, 1] == FILL_VALUE
        exterior_nodes = edge_node_connectivity[is_exterior]
        valid = np.full(len(vertices), True)
        valid[exterior_nodes.ravel()] = False
        valid = np.repeat(valid, ncol_per_row)
    else:
        valid = np.repeat(ncol_per_row >= 3, ncol_per_row)

    # Grab the centroids for all these faces.
    coo = node_face_connectivity.tocoo()
    node_i = coo.row[valid]
    face_i = coo.col[valid]
    voronoi_centroids = vertices[node_i]
    voronoi_vertices = centroids[face_i]
    # Compute the angle between voronoi centroids and vertices.
    x = voronoi_vertices[:, 0] - voronoi_centroids[:, 0]
    y = voronoi_vertices[:, 1] - voronoi_centroids[:, 1]
    angle = np.arctan2(y, x)
    # Now sort to create a counter clockwise oriented polygon.
    # Use a lexsort to make sure nodes are grouped together.
    order = np.lexsort((angle, node_i))
    j = face_i[order]

    if add_exterior:
        (
            vor_vertices,
            exterior_i,
            exterior_j,
            face_i,
            interpolation_map,
        ) = exterior_topology(
            edge_face_connectivity,
            edge_node_connectivity,
            node_face_connectivity,
            vertices,
            centroids,
            add_vertices,
            skip_concave,
        )
        offset = node_i.max() + 1 if len(node_i > 0) else 0
        i = np.concatenate([node_i, exterior_i + offset])
        j = np.concatenate([j, exterior_j])
    else:
        interpolation_map = None
        vor_vertices = centroids[np.unique(face_i)]
        face_i = np.arange(face_i.max() + 1)
        i = node_i
        j = renumber(j)

    face_node_connectivity = _create_face_node_connectivity(i, j)
    return vor_vertices, face_node_connectivity, face_i, interpolation_map



import xugrid as xu


class Network1d:
    def __init__(self, obj):
        # TODO: do not omit type check on grid!
        if isinstance(obj, (xu.UgridDataArray, xu.UgridDataset)):
            self.ugrid_topology = obj.grid
        elif isinstance(obj, xu.Ugrid1d):
            self.ugrid_topology = obj
        else:
            options = {"Ugrid1d", "UgridDataArray", "UgridDataset"}
            raise TypeError(
                f"Expected one of {options}, received: {type(obj).__name__}"
            )

    @property
    def ndim(self):
        return 1

    @property
    def dims(self):
        return (self.ugrid_topology.edge_dimension,)

    @property
    def shape(self):
        return (self.ugrid_topology.n_edge,)

    @property
    def size(self):
        return self.ugrid_topology.n_edge

    @property
    def length(self):
        return self.ugrid_topology.edge_length



import os
import warnings

import dask
import geopandas as gpd
import numpy as np
import pandas as pd
import pyproj
import pytest
import shapely
import xarray as xr

import xugrid


def GRID():
    VERTICES = np.array(
        [
            [0.0, 0.0],  # 0
            [1.0, 0.0],  # 1
            [2.0, 0.0],  # 2
            [0.0, 1.0],  # 3
            [1.0, 1.0],  # 4
            [2.0, 1.0],  # 5
            [1.0, 2.0],  # 6
        ]
    )
    FACES = np.array(
        [
            [0, 1, 4, 3],
            [1, 2, 5, 4],
            [3, 4, 6, -1],
            [4, 5, 6, -1],
        ]
    )
    return xugrid.Ugrid2d(
        node_x=VERTICES[:, 0],
        node_y=VERTICES[:, 1],
        fill_value=-1,
        face_node_connectivity=FACES,
    )


def DARRAY():
    return xr.DataArray(
        data=np.ones(GRID().n_face),
        dims=[GRID().face_dimension],
        name="a",
    )


def UGRID_DS():
    ds = GRID().to_dataset()
    ds["a"] = DARRAY()
    ds["b"] = DARRAY() * 2
    return ds


def ugrid1d_ds():
    xy = np.array(
        [
            [0.0, 0.0],
            [1.0, 1.0],
            [2.0, 2.0],
        ]
    )
    grid = xugrid.Ugrid1d(
        node_x=xy[:, 0],
        node_y=xy[:, 1],
        fill_value=-1,
        edge_node_connectivity=np.array([[0, 1], [1, 2]]),
    )
    ds = grid.to_dataset()
    ds["a1d"] = xr.DataArray([1.0, 2.0, 3.0], dims=[grid.node_dimension])
    ds["b1d"] = xr.DataArray([1.0, 2.0], dims=[grid.edge_dimension])
    return xugrid.UgridDataset(ds)


def test_properties():
    uda = xugrid.UgridDataArray(DARRAY(), GRID())
    uds = xugrid.UgridDataset(UGRID_DS())

    for item in (uda, uda.ugrid, uds, uds.ugrid):
        assert isinstance(getattr(item, "grid"), xugrid.Ugrid2d)
        grids = getattr(item, "grids")
        assert isinstance(grids, list)
        assert isinstance(grids[0], xugrid.Ugrid2d)

    assert isinstance(uda.obj, xr.DataArray)
    assert isinstance(uda.ugrid.obj, xr.DataArray)
    assert isinstance(uds.obj, xr.Dataset)
    assert isinstance(uds.ugrid.obj, xr.Dataset)


def test_xarray_property_setter():
    uda = xugrid.UgridDataArray(DARRAY(), GRID())
    uda.name = "new_name"
    assert uda.name == "new_name"


def test_init_errors():
    with pytest.raises(TypeError, match="obj must be xarray.DataArray"):
        xugrid.UgridDataArray(0, GRID())
    with pytest.raises(TypeError, match="grid must be Ugrid1d or Ugrid2d"):
        xugrid.UgridDataArray(DARRAY(), 0)

    with pytest.raises(ValueError, match="At least either obj or grids is required"):
        xugrid.UgridDataset()
    with pytest.raises(TypeError, match="obj must be xarray.Dataset"):
        xugrid.UgridDataset(0, GRID())
    with pytest.raises(TypeError, match="grid must be Ugrid1d or Ugrid2d"):
        xugrid.UgridDataset(xr.Dataset(), 0)


class TestUgridDataArray:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.uda = xugrid.UgridDataArray(DARRAY(), GRID())

    def test_properties(self):
        assert self.uda.ugrid.name == "mesh2d"
        assert self.uda.ugrid.names == ["mesh2d"]
        assert self.uda.ugrid.topology == {"mesh2d": self.uda.ugrid.grid}

    def test_init(self):
        assert isinstance(self.uda.ugrid.obj, xr.DataArray)
        assert isinstance(self.uda.ugrid.grid, xugrid.Ugrid2d)
        assert self.uda.grid.face_dimension in self.uda.coords

    def test_from_data(self):
        grid = self.uda.ugrid.grid
        uda = xugrid.UgridDataArray.from_data(np.zeros(grid.n_node), grid, facet="node")
        assert isinstance(uda, xugrid.UgridDataArray)

    def test_reinit_error(self):
        # Should not be able to initialize using a UgridDataArray.
        with pytest.raises(TypeError, match="obj must be xarray.DataArray"):
            xugrid.UgridDataArray(self.uda, GRID())

    def test_dunder_forward(self):
        assert isinstance(bool(self.uda[0]), bool)
        assert isinstance(int(self.uda[0]), int)
        assert isinstance(float(self.uda[0]), float)

    def test_close(self, tmp_path):
        path = tmp_path / "dataarray-closetest.nc"
        self.uda.ugrid.to_netcdf(path)
        back = xugrid.open_dataarray(path)
        back.close()
        os.remove(path)

    def test_repr(self):
        assert self.uda.__repr__() == self.uda.obj.__repr__()

    def test_getattr(self):
        # Get an attribute
        assert isinstance(self.uda.dims, tuple)
        # Only DataArrays should be returned as UgridDataArrays, other object
        # should remain untouched.
        assert self.uda.dims == self.uda.ugrid.obj.dims
        assert isinstance(self.uda.data, np.ndarray)
        # So are functions
        assert isinstance(self.uda.isnull(), xugrid.UgridDataArray)
        # obj should be accessible
        assert isinstance(self.uda.obj, xr.DataArray)

    def test_ugrid_accessor(self):
        assert isinstance(self.uda.ugrid, xugrid.UgridDataArrayAccessor)

    def test_rename(self):
        renamed = self.uda.ugrid.rename("renamed")
        assert "renamed_nFaces" in renamed.dims

    def test_unary_op(self):
        alltrue = self.uda.astype(bool)
        allfalse = alltrue.copy()
        allfalse[:] = False
        assert (~allfalse).all()
        assert isinstance(~allfalse, xugrid.UgridDataArray)

    def test_binary_op(self):
        alltrue = self.uda.astype(bool)
        allfalse = alltrue.copy()
        allfalse[:] = False
        assert isinstance(alltrue | allfalse, xugrid.UgridDataArray)
        assert (alltrue | allfalse).all()
        assert (alltrue ^ allfalse).all()
        assert not (alltrue & allfalse).any()
        # inplace op
        alltrue &= allfalse
        assert isinstance(alltrue, xugrid.UgridDataArray)
        assert not (alltrue).any()

    def test_math(self):
        actual = self.uda + 0
        assert isinstance(actual, xugrid.UgridDataArray)

    def test_np_ops(self):
        actual = np.abs(self.uda)
        assert isinstance(actual, xugrid.UgridDataArray)

    # Accessor tests
    def test_isel(self):
        actual = self.uda.isel({GRID().face_dimension: [0, 1]})
        assert isinstance(actual, xugrid.UgridDataArray)
        assert actual.shape == (2,)
        assert actual.ugrid.grid.n_face == 2

    def test_sel_points(self):
        with pytest.raises(ValueError, match="x and y must be 1d"):
            self.uda.ugrid.sel_points(x=[[0.0, 1.0]], y=[[0.0, 1.0]])
        with pytest.raises(ValueError, match="shape of x does not match shape of y"):
            self.uda.ugrid.sel_points(x=[0.0], y=[0.0, 1.0])
        actual = self.uda.ugrid.sel_points(x=[0.5, 0.5], y=[0.5, 1.25])
        assert isinstance(actual, xr.DataArray)
        assert actual.shape == (2,)

    def test_sel(self):
        # Ugrid2d already tests most
        # Orthogonal points
        x = [0.4, 0.8, 1.2]
        y = [0.25, 0.75]
        actual = self.uda.ugrid.sel(x=x, y=y)
        assert isinstance(actual, xr.DataArray)
        assert actual.shape == (6,)

        actual = self.uda.ugrid.sel(x=slice(0.4, 1.3, 0.4), y=0.25)
        assert isinstance(actual, xr.DataArray)
        assert actual.shape == (3,)

        actual = self.uda.ugrid.sel(x=slice(0, 1), y=slice(0, 2))
        assert isinstance(actual, xugrid.UgridDataArray)
        assert actual.shape == (2,)
        assert actual.ugrid.grid.n_face == 2

        actual = self.uda.ugrid.sel(x=slice(0, 1), y=slice(1, None))
        assert isinstance(actual, xugrid.UgridDataArray)
        assert actual.shape == (1,)
        assert actual.ugrid.grid.n_face == 1

    def test_intersect_line(self):
        p0 = (0.0, 0.0)
        p1 = (2.0, 2.0)
        actual = self.uda.ugrid.intersect_line(start=p0, end=p1)
        sqrt2 = np.sqrt(2.0)
        assert isinstance(actual, xr.DataArray)
        assert set(actual.dims) == {"mesh2d_nFaces"}
        assert np.allclose(actual["mesh2d_x"], [0.5, 1.25])
        assert np.allclose(actual["mesh2d_y"], [0.5, 1.25])
        assert np.allclose(actual["mesh2d_s"], [0.5 * sqrt2, 1.25 * sqrt2])

    def test_intersect_linestring(self):
        linestring = shapely.geometry.LineString(
            [
                [0.5, 0.5],
                [1.5, 0.5],
                [1.5, 1.5],
            ]
        )
        actual = self.uda.ugrid.intersect_linestring(linestring)
        assert isinstance(actual, xr.DataArray)
        assert set(actual.dims) == {"mesh2d_nFaces"}
        assert np.allclose(actual["mesh2d_x"], [0.75, 1.25, 1.5, 1.5])
        assert np.allclose(actual["mesh2d_y"], [0.5, 0.5, 0.75, 1.25])
        assert np.allclose(actual["mesh2d_s"], [0.25, 0.75, 1.25, 1.75])

    def test_rasterize(self):
        actual = self.uda.ugrid.rasterize(resolution=0.5)
        x = [0.25, 0.75, 1.25, 1.75]
        y = [1.75, 1.25, 0.75, 0.25]
        assert isinstance(actual, xr.DataArray)
        assert actual.shape == (4, 4)
        assert np.allclose(actual["x"], x)
        assert np.allclose(actual["y"], y)

        da = xr.DataArray(np.empty((4, 4)), {"y": y, "x": x}, ["y", "x"])
        actual = self.uda.ugrid.rasterize_like(other=da)
        assert isinstance(actual, xr.DataArray)
        assert actual.shape == (4, 4)
        assert np.allclose(actual["x"], x)
        assert np.allclose(actual["y"], y)

    def test_partitioning(self):
        partitions = self.uda.ugrid.partition(n_part=2)
        assert len(partitions) == 2
        for partition in partitions:
            assert isinstance(partition, xugrid.UgridDataArray)
            assert partition.name == self.uda.name

    def test_reindex_like(self):
        back = self.uda.ugrid.reindex_like(self.uda)
        assert isinstance(back, xugrid.UgridDataArray)
        back = self.uda.ugrid.reindex_like(self.uda.ugrid.grid)
        assert isinstance(back, xugrid.UgridDataArray)

    def test_crs(self):
        uda = self.uda.copy()
        crs = uda.ugrid.crs
        assert crs == {"mesh2d": None}

        uda.ugrid.set_crs(epsg=28992)
        assert uda.ugrid.crs == {"mesh2d": pyproj.CRS.from_epsg(28992)}

        result = uda.ugrid.to_crs(epsg=32631)
        assert uda is not result
        assert np.allclose(uda.values, result.values)
        assert uda.ugrid.crs == {"mesh2d": pyproj.CRS.from_epsg(28992)}
        assert result.ugrid.crs == {"mesh2d": pyproj.CRS.from_epsg(32631)}
        assert not np.array_equal(result.ugrid.grid.node_x, uda.ugrid.grid.node_x)
        assert not np.array_equal(result.ugrid.grid.node_y, uda.ugrid.grid.node_y)

    def test_crs_roundtrip(self):
        uda = self.uda.copy()
        uda.ugrid.set_crs(epsg=28992)
        ds = uda.ugrid.to_dataset()
        assert ds["a"].attrs == {"grid_mapping": "mesh2d_crs"}
        assert "mesh2d_crs" in ds.data_vars
        back = xugrid.UgridDataset(ds)
        assert back.ugrid.crs == {"mesh2d": pyproj.CRS.from_epsg(28992)}

    def test_is_geographic(self):
        uda = self.uda
        assert uda.grid.is_geographic is False
        assert uda.grid.is_projected is True

        uda.ugrid.set_crs(epsg=4326)
        assert uda.grid.is_geographic is True
        assert uda.grid.is_projected is False

        result = uda.ugrid.to_crs(epsg=28992)
        assert result.grid.is_geographic is False
        assert result.grid.is_projected is True

    def test_to_geodataframe(self):
        uda1 = self.uda.copy()
        uda1.ugrid.obj.name = None
        with pytest.raises(ValueError, match="unable to convert unnamed"):
            uda1.ugrid.to_geodataframe()
        uda2 = self.uda.copy()
        uda2.ugrid.obj.name = "test"
        uda2.ugrid.set_crs(epsg=28992)
        gdf = uda2.ugrid.to_geodataframe("mesh2d")
        assert isinstance(gdf, gpd.GeoDataFrame)
        assert (gdf.geometry.geom_type == "Polygon").all()
        assert gdf.crs.to_epsg() == 28992

    def test_binary_dilation(self):
        a = self.uda > 0
        actual = a.ugrid.binary_dilation()
        assert isinstance(actual, xugrid.UgridDataArray)

    def test_binary_erosion(self):
        a = self.uda > 0
        actual = a.ugrid.binary_erosion()
        assert isinstance(actual, xugrid.UgridDataArray)

    def test_connected_components(self):
        actual = self.uda.ugrid.connected_components()
        assert isinstance(actual, xugrid.UgridDataArray)
        assert np.allclose(actual, 0)

    def test_reverse_cuthill_mckee(self):
        actual = self.uda.ugrid.reverse_cuthill_mckee()
        assert isinstance(actual, xugrid.UgridDataArray)

    def test_laplace_interpolate(self):
        uda2 = self.uda.copy()
        uda2.obj[:-2] = np.nan
        actual = uda2.ugrid.laplace_interpolate(direct_solve=True)
        assert isinstance(actual, xugrid.UgridDataArray)
        assert np.allclose(actual, 1.0)

        actual = uda2.ugrid.laplace_interpolate(direct_solve=False)
        assert isinstance(actual, xugrid.UgridDataArray)
        assert np.allclose(actual, 1.0)

    def test_broadcasted_laplace_interpolate(self):
        uda2 = self.uda.copy()
        uda2.obj[:-2] = np.nan
        multiplier = xr.DataArray(
            np.ones((3, 2)),
            coords={"time": [0, 1, 2], "layer": [1, 2]},
            dims=("time", "layer"),
        )
        nd_uda2 = uda2 * multiplier
        actual = nd_uda2.ugrid.laplace_interpolate(direct_solve=True)
        assert isinstance(actual, xugrid.UgridDataArray)
        assert np.allclose(actual, 1.0)
        assert set(actual.dims) == set(nd_uda2.dims)

        actual = nd_uda2.ugrid.laplace_interpolate(direct_solve=False)
        assert isinstance(actual, xugrid.UgridDataArray)
        assert np.allclose(actual, 1.0)
        assert set(actual.dims) == set(nd_uda2.dims)

        # Test delayed evaluation too.
        nd_uda2 = uda2 * multiplier.chunk({"time": 1})
        actual = nd_uda2.ugrid.laplace_interpolate(direct_solve=True)
        assert isinstance(actual, xugrid.UgridDataArray)
        assert set(actual.dims) == set(nd_uda2.dims)
        assert isinstance(actual.data, dask.array.Array)

        actual = nd_uda2.ugrid.laplace_interpolate(direct_solve=False)
        assert isinstance(actual, xugrid.UgridDataArray)
        assert set(actual.dims) == set(nd_uda2.dims)
        assert isinstance(actual.data, dask.array.Array)

    def test_to_facets(self):
        face_da = self.uda
        grid = face_da.ugrid.grid

        with pytest.raises(ValueError, match="No conversion needed"):
            face_da.ugrid.to_face()
        with pytest.raises(ValueError, match="already exists"):
            face_da.ugrid.to_face(grid.face_dimension)

        node_da = face_da.ugrid.to_node()
        edge_da = face_da.ugrid.to_edge()
        assert isinstance(node_da, xugrid.UgridDataArray)
        assert isinstance(edge_da, xugrid.UgridDataArray)

        back1 = node_da.mean("nmax").ugrid.to_face()
        back2 = edge_da.mean("nmax").ugrid.to_face()
        cross1 = node_da.mean("nmax").ugrid.to_edge()
        cross2 = edge_da.mean("nmax").ugrid.to_node()

        assert isinstance(back1, xugrid.UgridDataArray)
        assert isinstance(back2, xugrid.UgridDataArray)
        assert isinstance(cross1, xugrid.UgridDataArray)
        assert isinstance(cross2, xugrid.UgridDataArray)

        assert node_da.dims == (grid.node_dimension, "nmax")
        assert edge_da.dims == (grid.edge_dimension, "nmax")
        assert back1.dims == (grid.face_dimension, "nmax")

        # Check fill values, should contain two NaNs for the fill value.
        assert back1[2:, -1].isnull().all()
        assert back1.isnull().sum() == 2

        # Multiple dimensions
        dim_multiplier = xr.DataArray([[1, 2, 3]], dims=("dim0", "dim1"))
        face_da_nd = face_da * dim_multiplier
        node_da_nd = face_da_nd.ugrid.to_node()
        assert node_da_nd.dims == (grid.node_dimension, "nmax", "dim0", "dim1")
        node_da_ndT = face_da_nd.transpose().ugrid.to_node()
        assert node_da_ndT.dims == ("dim1", "dim0", grid.node_dimension, "nmax")

    def test_to_dataset(self):
        uda2 = self.uda.copy()
        uda2.ugrid.obj.name = "test"
        actual = uda2.to_dataset()
        assert isinstance(actual, xugrid.UgridDataset)

    def test_ugrid_to_dataset(self):
        uda2 = self.uda.copy()
        uda2.ugrid.obj.name = "test"
        ds = uda2.ugrid.to_dataset(optional_attributes=True)
        assert "mesh2d_edge_nodes" in ds
        assert "mesh2d_face_nodes" in ds
        assert "mesh2d_face_edges" in ds
        assert "mesh2d_face_faces" in ds
        assert "mesh2d_edge_faces" in ds
        assert "mesh2d_boundary_nodes" in ds
        assert "mesh2d_face_x" in ds
        assert "mesh2d_face_y" in ds
        assert "mesh2d_edge_x" in ds
        assert "mesh2d_edge_y" in ds

    def test_to_netcdf(self, tmp_path):
        uda2 = self.uda.copy()
        uda2.ugrid.obj.name = "test"
        path = tmp_path / "uda-test.nc"
        uda2.ugrid.to_netcdf(path)
        assert path.exists()

    def test_to_zarr(self, tmp_path):
        uda2 = self.uda.copy()
        uda2.ugrid.obj.name = "test"
        path = tmp_path / "uda-test.zarr"
        uda2.ugrid.to_zarr(path)
        assert path.exists()

    def test_bounds(self):
        assert self.uda.ugrid.bounds == {"mesh2d": (0.0, 0.0, 2.0, 2.0)}

    def test_total_bounds(self):
        assert self.uda.ugrid.total_bounds == (0.0, 0.0, 2.0, 2.0)

    def test_assign_coords(self):
        with pytest.raises(ValueError):
            self.uda.ugrid.assign_edge_coords()
        with pytest.raises(ValueError):
            self.uda.ugrid.assign_node_coords()

        with_coords = self.uda.ugrid.assign_face_coords()
        assert "mesh2d_face_x" in with_coords.coords
        assert "mesh2d_face_y" in with_coords.coords

    def test_plot_with_chunks(self, tmp_path):
        time = xr.DataArray([0.0, 1.0, 2.0], coords={"time": [0, 1, 2]})
        uda = (self.uda * time).transpose()
        uda.name = "test"

        path = tmp_path / "test.nc"
        uda.ugrid.to_netcdf(path)
        back = xugrid.open_dataarray(path, chunks={"time": 1})
        primitive = back.isel(time=0).ugrid.plot()
        assert primitive is not None

    def test_plot_contourf_with_chunks(self, tmp_path):
        time = xr.DataArray([0.0, 1.0, 2.0], coords={"time": [0, 1, 2]})
        uda = (self.uda * time).transpose()
        uda.name = "test"

        path = tmp_path / "test.nc"
        uda.ugrid.to_netcdf(path)
        back = xugrid.open_dataarray(path, chunks={"time": 1})
        primitive = back.isel(time=0).ugrid.plot.contourf()
        assert primitive is not None


class TestUgridDataset:
    @pytest.fixture(autouse=True)
    def setup(self):
        ds = xr.Dataset()
        ds["a"] = DARRAY()
        ds["b"] = DARRAY() * 2
        self.uds = xugrid.UgridDataset(ds, GRID())

    def test_properties(self):
        assert self.uds.ugrid.name == "mesh2d"
        assert self.uds.ugrid.names == ["mesh2d"]
        assert self.uds.ugrid.topology == {"mesh2d": self.uds.ugrid.grid}

    def test_init(self):
        assert isinstance(self.uds.ugrid.obj, xr.Dataset)
        assert isinstance(self.uds.ugrid.grids[0], xugrid.Ugrid2d)
        # Try alternative initialization
        uds = xugrid.UgridDataset(grids=GRID())
        assert isinstance(uds, xugrid.UgridDataset)
        uds = xugrid.UgridDataset(grids=[GRID()])
        assert isinstance(uds, xugrid.UgridDataset)
        uds["a"] = DARRAY()
        assert "a" in uds.ugrid.obj

    def test_reinit_error(self):
        with pytest.raises(TypeError, match="obj must be xarray.Dataset"):
            xugrid.UgridDataset(self.uds, GRID())

    def test_init_from_dataset_only(self):
        uds = xugrid.UgridDataset(UGRID_DS())
        assert isinstance(uds, xugrid.UgridDataset)
        assert "a" in uds.ugrid.obj
        assert "b" in uds.ugrid.obj
        assert "mesh2d_face_nodes" in uds.ugrid.grids[0].to_dataset()
        assert "mesh2d_face_nodes" not in uds.ugrid.obj

    def test_repr(self):
        assert self.uds.__repr__() == self.uds.obj.__repr__()

    def test_close(self, tmp_path):
        path = tmp_path / "dataset-closetest.nc"
        self.uds.ugrid.to_netcdf(path)
        back = xugrid.open_dataset(path)
        back.close()
        os.remove(path)

    def test_getitem(self):
        assert "a" in self.uds
        assert "b" in self.uds
        assert isinstance(self.uds["a"], xugrid.UgridDataArray)
        assert isinstance(self.uds[["a", "b"]], xugrid.UgridDataset)

    def test_setitem(self):
        uds = self.uds.copy()
        # Set a UgridDataArray
        uds["b"] = self.uds["a"]
        assert (uds["b"].data == 1.0).all()
        # Set a scalar
        uds["a"] = 3.0
        assert (uds["a"].data == 3.0).all()

    def test_getattr(self):
        assert tuple(self.uds.dims) == ("mesh2d_nFaces",)
        assert isinstance(self.uds.a, xugrid.UgridDataArray)
        assert isinstance(self.uds.notnull(), xugrid.UgridDataset)
        # obj should be accessible
        assert isinstance(self.uds.obj, xr.Dataset)

    def test_unary_op(self):
        alltrue = self.uds.astype(bool)
        assert isinstance(~alltrue, xugrid.UgridDataset)

    def test_binary_op(self):
        alltrue = self.uds.astype(bool)
        assert isinstance(alltrue ^ alltrue, xugrid.UgridDataset)
        # inplace op
        alltrue &= alltrue
        assert isinstance(alltrue, xugrid.UgridDataset)

    def test_math(self):
        actual = self.uds + 0
        assert isinstance(actual, xugrid.UgridDataset)

    def test_ugrid_accessor(self):
        assert isinstance(self.uds.ugrid, xugrid.UgridDatasetAccessor)

    def test_rename(self):
        renamed = self.uds.ugrid.rename("renamed")
        assert "renamed_nFaces" in renamed.dims

        renamed = self.uds.ugrid.rename({"mesh2d": "renamed"})
        assert "renamed_nFaces" in renamed.dims

        # This name doesn't exist, shouldn't change
        renamed = self.uds.ugrid.rename({"mesh1d": "renamed"})
        assert "mesh2d_nFaces" in renamed.dims

        with pytest.raises(TypeError):
            self.uds.ugrid.rename(["mesh1d", "mesh2d"])

    def test_from_geodataframe(self):
        xy = np.array(
            [
                [0.0, 0.0],
                [1.0, 0.0],
                [1.0, 1.0],
                [0.0, 0.0],
            ]
        )
        polygon = shapely.polygons(xy)
        df = pd.DataFrame({"a": [1.0], "b": [2.0]})
        gdf = gpd.GeoDataFrame(df, geometry=[polygon])
        uds = xugrid.UgridDataset.from_geodataframe(gdf)
        assert isinstance(uds, xugrid.UgridDataset)
        assert "a" in uds
        assert "b" in uds

    # Accessor tests
    def test_isel(self):
        actual = self.uds.isel({GRID().face_dimension: [0, 1]})
        assert isinstance(actual, xugrid.UgridDataset)
        assert actual.ugrid.grids[0].n_face == 2
        assert actual["a"].shape == (2,)
        assert actual["b"].shape == (2,)

    def test_sel_points(self):
        with pytest.raises(ValueError, match="x and y must be 1d"):
            self.uds.ugrid.sel_points(x=[[0.0, 1.0]], y=[[0.0, 1.0]])
        with pytest.raises(ValueError, match="shape of x does not match shape of y"):
            self.uds.ugrid.sel_points(x=[0.0], y=[0.0, 1.0])
        actual = self.uds.ugrid.sel_points(x=[0.5, 0.5], y=[0.5, 1.25])
        assert isinstance(actual, xr.Dataset)
        assert actual["a"].shape == (2,)
        assert actual["b"].shape == (2,)

    def test_sel_points_multiple_dims(self):
        grid = self.uds.ugrid.grid
        ds = xr.Dataset(
            {
                "face_data": ((grid.face_dimension,), np.arange(grid.n_face)),
                "node_data": ((grid.node_dimension,), np.arange(grid.n_node)),
                "edge_data": ((grid.edge_dimension,), np.arange(grid.n_edge)),
            }
        )

        uds = xugrid.UgridDataset(ds, grid)
        x = [0.0, 1.0]
        y = [0.0, 1.0]
        actual = uds.ugrid.sel_points(x=x, y=y)

        dim = "mesh2d_points"
        expected = xr.Dataset(
            data_vars={
                "face_data": (dim, [0, 0]),
                "node_data": (dim, [0, 4]),
                "edge_data": (dim, [0, 3]),
            },
            coords={
                "mesh2d_nFaces": (dim, [0, 0]),
                "mesh2d_nEdges": (dim, [0, 3]),
                "mesh2d_nNodes": (dim, [0, 4]),
                "mesh2d_x": (dim, x),
                "mesh2d_y": (dim, y),
            },
        )
        assert actual.identical(expected)

    def test_sel(self):
        # Ugrid2d already tests most
        # Orthogonal points
        x = [0.4, 0.8, 1.2]
        y = [0.25, 0.75]
        actual = self.uds.ugrid.sel(x=x, y=y)
        assert isinstance(actual, xr.Dataset)
        assert actual["a"].shape == (6,)
        assert actual["b"].shape == (6,)

        actual = self.uds.ugrid.sel(x=slice(0.4, 1.3, 0.4), y=0.25)
        assert isinstance(actual, xr.Dataset)
        assert actual["a"].shape == (3,)
        assert actual["b"].shape == (3,)

        actual = self.uds.ugrid.sel(x=slice(0, 1), y=slice(0, 2))
        assert isinstance(actual, xugrid.UgridDataset)
        assert actual["a"].shape == (2,)
        assert actual["b"].shape == (2,)
        assert actual.ugrid.grids[0].n_face == 2

        actual = self.uds.ugrid.sel(x=slice(0, 1), y=slice(1, None))
        assert isinstance(actual, xugrid.UgridDataset)
        assert actual["a"].shape == (1,)
        assert actual["b"].shape == (1,)
        assert actual.ugrid.grids[0].n_face == 1

    def test_rasterize(self):
        actual = self.uds.ugrid.rasterize(resolution=0.5)
        x = [0.25, 0.75, 1.25, 1.75]
        y = [1.75, 1.25, 0.75, 0.25]
        assert isinstance(actual, xr.Dataset)
        assert actual["a"].shape == (4, 4)
        assert actual["b"].shape == (4, 4)
        assert np.allclose(actual["x"], x)
        assert np.allclose(actual["y"], y)

        da = xr.DataArray(np.empty((4, 4)), {"y": y, "x": x}, ["y", "x"])
        actual = self.uds.ugrid.rasterize_like(other=da)
        assert isinstance(actual, xr.Dataset)
        assert actual["a"].shape == (4, 4)
        assert actual["b"].shape == (4, 4)
        assert np.allclose(actual["x"], x)
        assert np.allclose(actual["y"], y)

    def test_intersect_line(self):
        p0 = (0.0, 0.0)
        p1 = (2.0, 2.0)
        actual = self.uds.ugrid.intersect_line(start=p0, end=p1)
        sqrt2 = np.sqrt(2.0)
        assert isinstance(actual, xr.Dataset)
        assert actual.sizes == {"mesh2d_nFaces": 2}
        assert np.allclose(actual["mesh2d_x"], [0.5, 1.25])
        assert np.allclose(actual["mesh2d_y"], [0.5, 1.25])
        assert np.allclose(actual["mesh2d_s"], [0.5 * sqrt2, 1.25 * sqrt2])
        assert "a" in actual
        assert "b" in actual

    def test_intersect_linestring(self):
        linestring = shapely.geometry.LineString(
            [
                [0.5, 0.5],
                [1.5, 0.5],
                [1.5, 1.5],
            ]
        )
        actual = self.uds.ugrid.intersect_linestring(linestring)
        assert isinstance(actual, xr.Dataset)
        assert actual.sizes == {"mesh2d_nFaces": 4}
        assert np.allclose(actual["mesh2d_x"], [0.75, 1.25, 1.5, 1.5])
        assert np.allclose(actual["mesh2d_y"], [0.5, 0.5, 0.75, 1.25])
        assert np.allclose(actual["mesh2d_s"], [0.25, 0.75, 1.25, 1.75])
        assert "a" in actual
        assert "b" in actual

    def test_partitioning(self):
        partitions = self.uds.ugrid.partition(n_part=2)
        assert len(partitions) == 2
        for partition in partitions:
            assert isinstance(partition, xugrid.UgridDataset)
            assert "a" in partition
            assert "b" in partition

    def test_reindex_like(self):
        back = self.uds.ugrid.reindex_like(self.uds)
        assert isinstance(back, xugrid.UgridDataset)
        back = self.uds.ugrid.reindex_like(self.uds.ugrid.grid)
        assert isinstance(back, xugrid.UgridDataset)

    def test_crs(self):
        uds = self.uds.copy()
        crs = uds.ugrid.crs
        assert crs == {"mesh2d": None}

        with pytest.raises(ValueError, match="grid not found"):
            uds.ugrid.set_crs(epsg=28992, topology="grid")

        uds.ugrid.set_crs(epsg=28992, topology="mesh2d")
        assert uds.ugrid.crs == {"mesh2d": pyproj.CRS.from_epsg(28992)}

        uds.ugrid.set_crs(epsg=28992)
        assert uds.ugrid.crs == {"mesh2d": pyproj.CRS.from_epsg(28992)}

        with pytest.raises(ValueError, match="grid not found"):
            uds.ugrid.to_crs(epsg=32631, topology="grid")

        result = uds.ugrid.to_crs(epsg=32631, topology="mesh2d")
        assert uds is not result
        assert uds.ugrid.crs == {"mesh2d": pyproj.CRS.from_epsg(28992)}
        assert result.ugrid.crs == {"mesh2d": pyproj.CRS.from_epsg(32631)}

    def test_crs_from_minimal(self):
        # Grid mapping on a single variable
        # Just epsg
        ds = UGRID_DS()
        ds["a"].attrs["grid_mapping"] = "projected_crs"
        ds["projected_crs"] = xr.Variable(dims=(), data=0, attrs={"epsg": 28992})
        uds = xugrid.UgridDataset(ds)
        assert uds.ugrid.crs == {"mesh2d": pyproj.CRS.from_epsg(28992)}

    def test_crs_roundtrip(self):
        uds = self.uds.copy()
        uds.ugrid.set_crs(epsg=28992, topology="mesh2d")
        ds = uds.ugrid.to_dataset()
        expected = {"grid_mapping": "mesh2d_crs"}
        assert ds["a"].attrs == expected
        assert ds["b"].attrs == expected
        assert "mesh2d_crs" in ds.data_vars
        back = xugrid.UgridDataset(ds)
        assert back.ugrid.crs == {"mesh2d": pyproj.CRS.from_epsg(28992)}
        # grid_mapping should still be present on ds, should not have been popped.
        assert "grid_mapping" in ds["a"].attrs
        assert "grid_mapping" in ds["b"].attrs

    def test_assign_coords(self):
        with_coords = (
            self.uds.ugrid.assign_edge_coords()
            .ugrid.assign_node_coords()
            .ugrid.assign_face_coords()
        )
        assert "mesh2d_node_x" in with_coords.coords
        assert "mesh2d_node_y" in with_coords.coords
        assert "mesh2d_edge_x" in with_coords.coords
        assert "mesh2d_edge_y" in with_coords.coords
        assert "mesh2d_face_x" in with_coords.coords
        assert "mesh2d_face_y" in with_coords.coords

    def test_to_geodataframe(self):
        gdf = self.uds.ugrid.to_geodataframe()
        assert isinstance(gdf, gpd.GeoDataFrame)
        assert (gdf.geometry.geom_type == "Polygon").all()

        # Now test with an empty dataset
        grid = self.uds.grid
        empty = xugrid.UgridDataset(grid.to_dataset())
        empty.ugrid.set_crs(epsg=28992)
        gdf = empty.ugrid.to_geodataframe()
        assert isinstance(gdf, gpd.GeoDataFrame)
        assert gdf.crs.to_epsg() == 28992

    def test_bounds(self):
        assert self.uds.ugrid.bounds == {"mesh2d": (0.0, 0.0, 2.0, 2.0)}

    def test_total_bounds(self):
        assert self.uds.ugrid.total_bounds == (0.0, 0.0, 2.0, 2.0)


class TestDatasetOptionalCoordinates:
    """Test whether all coordinates, not just nodes, are kept in sync."""

    @pytest.fixture(autouse=True)
    def setup(self):
        self.grid = GRID()
        self.ds = self.grid.to_dataset(optional_attributes=True)
        self.ds["a"] = DARRAY()
        self.X = ("mesh2d_node_x", "mesh2d_edge_x", "mesh2d_face_x")
        self.Y = ("mesh2d_node_y", "mesh2d_edge_y", "mesh2d_face_y")

    def test_indexes(self):
        uds = xugrid.UgridDataset(self.ds)
        expected = {
            "node_x": "mesh2d_node_x",
            "node_y": "mesh2d_node_y",
            "edge_x": "mesh2d_edge_x",
            "edge_y": "mesh2d_edge_y",
            "face_x": "mesh2d_face_x",
            "face_y": "mesh2d_face_y",
        }
        assert uds.grid._indexes == expected

    def test_dataset_set_crs(self):
        uds = xugrid.UgridDataset(self.ds)
        for x in self.X:
            assert uds[x].attrs["standard_name"] == "projection_x_coordinate"
        for x in self.Y:
            assert uds[x].attrs["standard_name"] == "projection_y_coordinate"
        uds.ugrid.set_crs(epsg=28992)
        for x in self.X:
            assert uds[x].attrs["standard_name"] == "projection_x_coordinate"
        for x in self.Y:
            assert uds[x].attrs["standard_name"] == "projection_y_coordinate"
        uds.ugrid.set_crs(epsg=4326, allow_override=True)
        for x in self.X:
            assert uds[x].attrs["standard_name"] == "longitude"
        for x in self.Y:
            assert uds[x].attrs["standard_name"] == "latitude"

        back = uds.ugrid.to_dataset()
        for x in self.X:
            assert back[x].attrs["standard_name"] == "longitude"
        for x in self.Y:
            assert back[x].attrs["standard_name"] == "latitude"

    def test_dataset_to_crs(self):
        uds = xugrid.UgridDataset(self.ds)
        uds.ugrid.set_crs(epsg=28992)
        for x in self.X:
            assert uds[x].attrs["standard_name"] == "projection_x_coordinate"
        for x in self.Y:
            assert uds[x].attrs["standard_name"] == "projection_y_coordinate"
        new = uds.ugrid.to_crs(epsg=4326)
        for x in self.X:
            assert new[x].attrs["standard_name"] == "longitude"
        for x in self.Y:
            assert new[x].attrs["standard_name"] == "latitude"

        def is_different(a, b, name):
            return (a[name] != b[name]).all()

        names = (
            "mesh2d_node_x",
            "mesh2d_node_y",
            "mesh2d_edge_x",
            "mesh2d_edge_y",
            "mesh2d_face_x",
            "mesh2d_face_y",
        )
        for name in names:
            assert is_different(uds, new, name)

        back = new.ugrid.to_dataset()
        for x in self.X:
            assert back[x].attrs["standard_name"] == "longitude"
        for x in self.Y:
            assert back[x].attrs["standard_name"] == "latitude"

    def test_dropped_grid_mapping(self):
        grid = self.grid.copy()
        grid.set_crs(epsg=28992)
        ds = grid.to_dataset()
        ds["a"] = DARRAY()
        ds["a"].attrs["grid_mapping"] = "mesh2d_crs"
        assert "mesh2d_crs" in ds
        uds = xugrid.UgridDataset(ds)
        assert "mesh2d_crs" not in uds
        assert "grid_mapping" not in uds["a"].attrs
        back = uds.ugrid.to_dataset()
        assert "mesh2d_crs" in back
        assert "grid_mapping" in back["a"].attrs

    def test_dataarray_set_crs(self):
        uds = xugrid.UgridDataset(self.ds)
        uda = uds["a"]
        uda.ugrid.set_crs(epsg=4326)
        assert uda["mesh2d_face_x"].attrs["standard_name"] == "longitude"
        assert uda["mesh2d_face_y"].attrs["standard_name"] == "latitude"

        back = uda.ugrid.to_dataset()
        for x in self.X:
            assert back[x].attrs["standard_name"] == "longitude"
        for x in self.Y:
            assert back[x].attrs["standard_name"] == "latitude"

    def test_dataarray_to_crs(self):
        uds = xugrid.UgridDataset(self.ds)
        uda = uds["a"]
        uda.ugrid.set_crs(epsg=28992)
        new = uda.ugrid.to_crs(epsg=4326)
        assert new["mesh2d_face_x"].attrs["standard_name"] == "longitude"
        assert new["mesh2d_face_y"].attrs["standard_name"] == "latitude"
        back = new.ugrid.to_dataset()
        assert back["mesh2d_face_x"].attrs["standard_name"] == "longitude"
        assert back["mesh2d_face_y"].attrs["standard_name"] == "latitude"
        # These have been invalidated and should not longer be present.
        assert "mesh2d_edge_x" not in back
        assert "mesh2d_edge_y" not in back


class TestMultiTopologyUgridDataset:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.uds = xugrid.UgridDataset(grids=GRID())
        uda = xugrid.UgridDataArray(DARRAY(), GRID())
        self.uds["a"] = uda

        xy = np.array(
            [
                [0.0, 0.0],
                [1.0, 1.0],
                [2.0, 2.0],
            ]
        )
        grid = xugrid.Ugrid1d(
            node_x=xy[:, 0],
            node_y=xy[:, 1],
            fill_value=-1,
            edge_node_connectivity=np.array([[0, 1], [1, 2]]),
        )
        self.uds["b"] = xugrid.UgridDataArray(
            xr.DataArray(np.ones(grid.n_node), dims=[grid.node_dimension]),
            grid,
        )
        self.uds["c"] = xugrid.UgridDataArray(
            xr.DataArray(np.ones(grid.n_edge), dims=[grid.edge_dimension]),
            grid,
        )

    def test_grid_membership(self):
        assert len(self.uds.grids) == 2

    def test_grid_accessor__error(self):
        with pytest.raises(TypeError):
            self.uds.ugrid.grid

        with pytest.raises(TypeError):
            self.uds.grid

    def test_multi_topology_sel(self):
        result = self.uds.ugrid.sel(x=slice(-10, 10), y=slice(-10, 10))
        # Ensure both grids are still present
        assert len(result.ugrid.grids) == 2

    def test_multi_topology_isel(self):
        grid0, grid1 = self.uds.grids
        result0 = self.uds.isel({grid0.face_dimension: [0, 1]})
        assert len(result0.ugrid.grids) == 2

        result1 = self.uds.isel({grid1.edge_dimension: [0, 1]})
        assert len(result1.ugrid.grids) == 2

    def test_reindex_like(self):
        back = self.uds.ugrid.reindex_like(self.uds)
        assert isinstance(back, xugrid.UgridDataset)


class TestFromStructured:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.da1d = xr.DataArray(
            [0.0, 1.0, 2.0, 3.0], {"x": [2.0, 4.0, 6.0, 8.0]}, ["x"]
        )
        self.da2d = xr.DataArray(
            data=np.arange(2 * 3 * 4).reshape((2, 3, 4)),
            coords={"layer": [1, 2], "y": [5.0, 10.0, 15.0], "x": [2.0, 4.0, 6.0, 8.0]},
            dims=["layer", "y", "x"],
            name="grid",
        )
        self.da_coords2d = xr.DataArray(
            data=[[0, 1], [2, 3]],
            coords={
                "yc": (("y", "x"), [[12.0, 11.0], [12.0, 11.0]]),
                "xc": (("y", "x"), [[10.0, 12.0], [10.0, 12.0]]),
            },
            dims=("y", "x"),
        )
        self.ds = xr.Dataset(
            {
                "a": self.da2d,
                "b": self.da1d,
                "c": 1.0,
            }
        )

    def test_future_warnings(self):
        with pytest.warns(FutureWarning):
            xugrid.UgridDataArray.from_structured(self.da2d)
        with pytest.warns(FutureWarning):
            xugrid.UgridDataset.from_structured(self.ds)

    def test_error_1d(self):
        with pytest.raises(
            ValueError, match="DataArray must have at least two spatial dimensions"
        ):
            xugrid.UgridDataArray.from_structured2d(self.da1d)

    def test_error_x_xor_y(self):
        with pytest.raises(ValueError, match="Provide both x and y, or neither."):
            xugrid.UgridDataArray.from_structured2d(self.da2d, x="this")

    def test_missing_xy(self):
        with pytest.raises(ValueError, match="Coordinates xc and yc are not present."):
            xugrid.UgridDataArray.from_structured2d(self.da2d, x="xc", y="yc")

    def test_from_dataarray(self):
        uda = xugrid.UgridDataArray.from_structured2d(self.da2d)
        assert isinstance(uda, xugrid.UgridDataArray)
        assert uda.name == "grid"
        assert uda.dims == ("layer", "mesh2d_nFaces")
        assert uda.shape == (2, 12)
        assert np.allclose(uda.ugrid.sel(x=2.0, y=5.0), [[0], [12]])
        # Check whether flipping the y-axis doesn't cause any problems
        flipped = self.da2d.isel(y=slice(None, None, -1))
        uda = xugrid.UgridDataArray.from_structured2d(flipped)
        assert np.allclose(uda.ugrid.sel(x=2.0, y=5.0), [[0], [12]])

        # And test transposed.
        daT = self.da2d.transpose()
        uda = xugrid.UgridDataArray.from_structured2d(daT)
        assert isinstance(uda, xugrid.UgridDataArray)
        assert uda.name == "grid"
        assert uda.dims == ("layer", "mesh2d_nFaces")
        assert uda.shape == (2, 12)
        assert np.allclose(uda.ugrid.sel(x=2.0, y=5.0), [[0], [12]])

    def test_from_multicoord(self):
        uda = xugrid.UgridDataArray.from_structured2d(self.da_coords2d)
        assert isinstance(uda, xugrid.UgridDataArray)
        assert np.array_equal(np.unique(uda.ugrid.grid.node_x), [-0.5, 0.5, 1.5])
        assert np.array_equal(uda.data, [0, 1, 2, 3])

        uda = xugrid.UgridDataArray.from_structured2d(self.da_coords2d, x="xc", y="yc")
        assert isinstance(uda, xugrid.UgridDataArray)
        assert np.array_equal(np.unique(uda.ugrid.grid.node_x), [9.0, 11.0, 13.0])
        assert np.array_equal(uda.data, [0, 1, 2, 3])

    def test_from_dataset(self):
        uds = xugrid.UgridDataset.from_structured2d(self.ds)
        assert isinstance(uds, xugrid.UgridDataset)
        assert set(uds.data_vars) == {"a", "b", "c"}
        assert uds["a"].dims == ("layer", "mesh2d_nFaces")
        assert uds["b"].dims == ("x",)
        assert uds["c"].dims == ()
        uda = uds["a"]
        assert uda.shape == (2, 12)
        assert np.allclose(uda.ugrid.sel(x=2.0, y=5.0), [[0], [12]])

    def test_from_multicoord_dataset(self):
        ds = self.ds.copy()
        da = self.da_coords2d.rename({"x": "x1", "y": "y1"})
        ds["d"] = da
        # Unspecified: it'll only infer x and y.
        uds = xugrid.UgridDataset.from_structured2d(ds)
        assert isinstance(uds, xugrid.UgridDataset)
        assert uds["a"].dims == ("layer", "mesh2d_nFaces")
        assert uds["d"].dims == ("y1", "x1")
        assert len(uds.ugrid.grids) == 1
        # Now specify separate topologies.
        uds = xugrid.UgridDataset.from_structured2d(
            ds, {"mesh2d_0": ("x", "y"), "mesh2d_1": ("xc", "yc")}
        )
        assert isinstance(uds, xugrid.UgridDataset)
        assert uds["a"].dims == ("layer", "mesh2d_0_nFaces")
        assert uds["b"].dims == ("x",)
        assert uds["c"].dims == ()
        assert uds["d"].dims == ("mesh2d_1_nFaces",)
        assert len(uds.ugrid.grids) == 2

    def test_from_bounds(self):
        # Construct bounds from ugrid form, it's easier...
        uda = xugrid.UgridDataArray.from_structured2d(self.da2d)
        grid = uda.ugrid.grid
        bounds_x = xr.DataArray(
            grid.face_node_coordinates[..., 0].reshape(3, 4, 4),
            dims=("y", "x", "bound"),
        )
        bounds_y = xr.DataArray(
            grid.face_node_coordinates[..., 1].reshape(3, 4, 4),
            dims=("y", "x", "bound"),
        )
        uda2 = xugrid.UgridDataArray.from_structured2d(
            self.da2d, "x", "y", bounds_x, bounds_y
        )
        assert uda.identical(uda2)
        # Try inconsistent dim order on bounds:
        uda3 = xugrid.UgridDataArray.from_structured2d(
            self.da2d, "x", "y", bounds_x.transpose(), bounds_y
        )
        assert uda.identical(uda3)

        with pytest.raises(ValueError, match="x and y must be provided for bounds"):
            xugrid.UgridDataArray.from_structured2d(
                self.da2d, x_bounds=bounds_x, y_bounds=bounds_y
            )

        # Now dataset
        ds = self.ds.copy()
        ds["grid_x"] = bounds_x
        ds["grid_y"] = bounds_y
        uds = xugrid.UgridDataset.from_structured2d(
            ds,
            topology={
                "mesh2d": {
                    "x": "x",
                    "y": "y",
                    "x_bounds": "grid_x",
                    "y_bounds": "grid_y",
                }
            },
        )
        assert set(uds.data_vars) == {"a", "b", "c"}
        assert uds["a"].dims == ("layer", "mesh2d_nFaces")
        assert uds["b"].dims == ("x",)
        assert uds["c"].dims == ()

        with pytest.raises(TypeError):
            xugrid.UgridDataset.from_structured2d(
                ds, topology={"x_bounds": "grid_x", "y_bounds": "grid_y"}
            )

        with pytest.raises(ValueError, match="x and y must be provided for bounds"):
            xugrid.UgridDataset.from_structured2d(
                ds, topology={"mesh2d": {"x_bounds": "grid_x", "y_bounds": "grid_y"}}
            )


def test_multiple_coordinates():
    grid = GRID()
    ds = UGRID_DS()
    attrs = ds["mesh2d"].attrs
    attrs["node_coordinates"] += " mesh2d_node_lon mesh2d_node_lat"
    ds = ds.assign_coords(
        mesh2d_node_lon=(grid.node_dimension, np.arange(grid.n_node)),
        mesh2d_node_lat=(grid.node_dimension, np.arange(grid.n_node)),
    )
    ds["mesh2d_node_lon"].attrs["standard_name"] = "longitude"
    ds["mesh2d_node_lat"].attrs["standard_name"] = "latitude"
    assert ds.ugrid_roles.coordinates == {
        "mesh2d": {
            "node_coordinates": (
                ["mesh2d_node_x", "mesh2d_node_lon"],
                ["mesh2d_node_y", "mesh2d_node_lat"],
            )
        }
    }

    # Make sure everything goes right when subsetting: tests whether all
    # attributes and grid indexes are propagated to the new grid object.
    with pytest.warns(UserWarning) as record:
        uds = xugrid.UgridDataset(ds)

    messages = [str(w.message) for w in record]
    assert any("Inconsistent standard_names across coordinates" in m for m in messages)
    assert any("No CRS or recognizable standard_name" in m for m in messages)

    subset = uds.isel({grid.face_dimension: [0, 1]})
    assert isinstance(subset, xugrid.UgridDataset)
    subset_ds = uds.ugrid.to_dataset()
    assert "mesh2d_node_x" in subset_ds
    assert "mesh2d_node_y" in subset_ds
    assert "mesh2d_node_lon" in subset_ds
    assert "mesh2d_node_lat" in subset_ds
    assert subset_ds["mesh2d"].attrs["node_coordinates"] == attrs["node_coordinates"]


def test_ugrid_to_dataset():
    uds = xugrid.UgridDataset(UGRID_DS())
    assert uds.ugrid.to_dataset() == UGRID_DS()

    ds = uds.ugrid.to_dataset(optional_attributes=True)
    assert "mesh2d_edge_nodes" in ds
    assert "mesh2d_face_nodes" in ds
    assert "mesh2d_face_edges" in ds
    assert "mesh2d_face_faces" in ds
    assert "mesh2d_edge_faces" in ds
    assert "mesh2d_boundary_nodes" in ds
    assert "mesh2d_face_x" in ds
    assert "mesh2d_face_y" in ds
    assert "mesh2d_edge_x" in ds
    assert "mesh2d_edge_y" in ds


def test_open_dataset(tmp_path):
    path = tmp_path / "ugrid-dataset.nc"
    uds = xugrid.UgridDataset(UGRID_DS())
    uds.ugrid.to_netcdf(path)

    back = xugrid.open_dataset(path)
    assert isinstance(back, xugrid.UgridDataset)
    assert "b" in back
    assert "mesh2d_face_nodes" in back.ugrid.grids[0].to_dataset()
    assert "mesh2d_face_nodes" not in back.ugrid.obj

    path = tmp_path / "no-topology-dataset.nc"
    uds.to_netcdf(path)
    with pytest.raises(
        ValueError, match="The file or object does not contain UGRID conventions data."
    ):
        xugrid.open_dataset(path)


def test_load_dataset(tmp_path):
    path = tmp_path / "ugrid-dataset.nc"
    uds = xugrid.UgridDataset(UGRID_DS())
    uds.ugrid.to_netcdf(path)

    back = xugrid.load_dataset(path)
    assert isinstance(back, xugrid.UgridDataset)
    assert "b" in back
    assert "mesh2d_face_nodes" in back.ugrid.grids[0].to_dataset()
    assert "mesh2d_face_nodes" not in back.ugrid.obj


def test_open_dataset_cast_invalid(tmp_path):
    grid = GRID()
    vorgrid = grid.tesselate_centroidal_voronoi()
    path = tmp_path / "voronoi-grid.nc"
    vorgrid.to_dataset().to_netcdf(path)

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        xugrid.open_dataset(path)


def test_open_dataarray_roundtrip(tmp_path):
    path = tmp_path / "ugrid-dataset.nc"
    uds = xugrid.UgridDataset(UGRID_DS())
    uds.ugrid.to_netcdf(path)
    with pytest.raises(ValueError, match="The file or object contains more than one"):
        xugrid.open_dataarray(path)

    path = tmp_path / "ugrid-dataarray.nc"
    uds["a"].ugrid.to_netcdf(path)
    back = xugrid.open_dataarray(path)
    assert isinstance(back, xugrid.UgridDataArray)
    assert back.name == "a"


def test_open_dataarray_ugrid_errors(tmp_path):
    uds = ugrid1d_ds()
    path = tmp_path / "no-topology-dataarray.nc"
    uds.to_netcdf(path)
    with pytest.raises(
        ValueError, match="The file or object does not contain UGRID conventions data."
    ):
        xugrid.open_dataarray(path)


def test_load_dataarray_roundtrip(tmp_path):
    path = tmp_path / "ugrid-dataset.nc"
    uds = xugrid.UgridDataset(UGRID_DS())
    uds.ugrid.to_netcdf(path)
    with pytest.raises(ValueError, match="The file or object contains more than one"):
        xugrid.load_dataarray(path)

    path = tmp_path / "ugrid-dataarray.nc"
    uds["a"].ugrid.to_netcdf(path)
    back = xugrid.load_dataarray(path)
    assert isinstance(back, xugrid.UgridDataArray)
    assert back.name == "a"


def test_open_mfdataset(tmp_path):
    path1 = tmp_path / "ugrid-dataset_1.nc"
    path2 = tmp_path / "ugrid-dataset_2.nc"
    uds = xugrid.UgridDataset(UGRID_DS())
    uda1 = uds["a"].expand_dims(dim="layer")
    uda2 = uds["a"].expand_dims(dim="layer")
    uda1 = uda1.assign_coords(layer=[1])
    uda2 = uda1.assign_coords(layer=[2])
    uda1.ugrid.to_netcdf(path1)
    uda2.ugrid.to_netcdf(path2)
    back = xugrid.open_mfdataset([path1, path2])
    assert isinstance(back, xugrid.UgridDataset)
    assert "a" in back
    assert tuple(back["a"].dims) == ("layer", "mesh2d_nFaces")

    with pytest.raises(ValueError, match="data_vars kwargs is not supported"):
        back = xugrid.open_mfdataset([path1, path2], data_vars="all")


def test_keep_attrs():
    uds = xugrid.UgridDataset(UGRID_DS())
    uds.attrs["date_created"] = "today"
    ds = uds.ugrid.to_dataset()
    assert ds.attrs["date_created"] == "today"

    uds = ugrid1d_ds()
    uds.attrs["date_created"] = "today"
    ds = uds.ugrid.to_dataset()
    assert ds.attrs["date_created"] == "today"


def test_zarr_roundtrip(tmp_path):
    path = tmp_path / "ugrid-dataset.zarr"
    uds = xugrid.UgridDataset(UGRID_DS())
    uds.ugrid.to_zarr(path)

    back = xugrid.open_zarr(path)
    assert isinstance(back, xugrid.UgridDataset)
    assert "a" in back
    assert "b" in back
    assert "mesh2d_face_nodes" in back.ugrid.grids[0].to_dataset()
    assert "mesh2d_face_nodes" not in back.ugrid.obj


def test_func_like():
    uds = xugrid.UgridDataset(UGRID_DS())

    fullda = xugrid.full_like(uds["a"], 2)
    assert isinstance(fullda, xugrid.UgridDataArray)
    assert (fullda == 2).all()
    # Topology should be untouched
    assert fullda.ugrid.grid.to_dataset() == uds.ugrid.grids[0].to_dataset()

    fullds = xugrid.full_like(uds, 2)
    assert isinstance(fullds, xugrid.UgridDataset)
    assert (fullds["a"] == 2).all()
    assert (fullds["b"] == 2).all()
    assert fullds.ugrid.grids[0].to_dataset == uds.ugrid.grids[0].to_dataset()

    fullda = xugrid.zeros_like(uds["a"])
    assert isinstance(fullda, xugrid.UgridDataArray)
    assert (fullda == 0).all()

    fullda = xugrid.ones_like(uds["a"])
    assert isinstance(fullda, xugrid.UgridDataArray)
    assert (fullda == 1).all()


def test_concat():
    uds = xugrid.UgridDataset(UGRID_DS())
    uda = uds["a"]
    uda1 = uda.assign_coords(layer=1)
    uda2 = uda.assign_coords(layer=2)
    result = xugrid.concat([uda1, uda2], dim="layer")
    assert result.dims == ("layer", "mesh2d_nFaces")
    assert np.array_equal(result["layer"], [1, 2])

    uds1d = ugrid1d_ds()
    uda3 = uds1d["a1d"].assign_coords(layer=2)
    with pytest.raises(ValueError, match="All UgridDataArrays must have the same grid"):
        xugrid.concat([uda1, uda3], dim="layer")

    # test issue 206 resolved
    # https://github.com/Deltares/xugrid/issues/206
    result = xugrid.concat([uda1, uda2.copy()], dim="foo")
    assert len(result.grids) == 1


def test_multiple_topology_errors():
    # Create a dataset with two UGRID topologies:
    uds = ugrid1d_ds()
    uds["a"] = xugrid.UgridDataset(UGRID_DS())["a"]

    with pytest.raises(TypeError, match="Can only access grid topology"):
        uds.ugrid.grid

    with pytest.raises(TypeError, match="Can only access grid name"):
        uds.ugrid.name

    with pytest.raises(TypeError, match="Can only rename with a single name"):
        uds.ugrid.rename("renamed")


def test_merge():
    uds2d = xugrid.UgridDataset(UGRID_DS())
    uds1d = ugrid1d_ds()
    merged = xugrid.merge([uds2d, uds1d])
    assert isinstance(merged, xugrid.UgridDataset)
    assert len(merged.grids) == 2


def get_ugrid_fillvaluem999_startindex1_ds():
    """
    Return a minimal dataset with a specific fill value.

    This is a very minimal but comparable dataset to Grevelingen_0002_map.nc
    (FM output).

    * It contains both triangles and squares.
    * The fillvalue of the connectivity arrays is -999
    * The start_index of the connectivity arrays is 1
    """

    ds2 = xr.Dataset()

    mesh2d_attrs = {
        "cf_role": "mesh_topology",
        "topology_dimension": 2,
        "node_dimension": "nmesh2d_node",
        "edge_dimension": "nmesh2d_edge",
        "face_dimension": "nmesh2d_face",
        "max_face_nodes_dimension": "max_nmesh2d_face_nodes",
        "face_node_connectivity": "mesh2d_face_nodes",
        "edge_node_connectivity": "mesh2d_edge_nodes",
        "node_coordinates": "mesh2d_node_x mesh2d_node_y",
        "name": "mesh2d",
    }
    ds2["mesh2d"] = xr.DataArray(np.array(0, dtype=int), attrs=mesh2d_attrs)

    node_x = np.array(
        [
            48231.65428822,
            48264.81400401,
            48350.0,
            48450.0,
            48235.96727817,
            48287.80875187,
            48306.85396605,
            48400.0,
            48500.0,
            48273.38390534,
            48450.0,
            48350.77800678,
            48342.73889736,
        ]
    )
    node_x_attrs = {
        "units": "m",
        "standard_name": "projection_x_coordinate",
        "long_name": "x-coordinate of mesh nodes",
        "mesh": "mesh2d",
        "location": "node",
    }
    ds2["mesh2d_node_x"] = xr.DataArray(
        node_x, dims=(mesh2d_attrs["node_dimension"]), attrs=node_x_attrs
    )

    node_y = np.array(
        [
            419541.94243367,
            419605.7455447,
            419605.94894,
            419605.94894,
            419454.05602638,
            419488.96173469,
            419552.39342675,
            419519.3464,
            419519.3464,
            419418.34074563,
            419432.74386,
            419447.63359856,
            419378.75075546,
        ]
    )
    node_y_attrs = {
        "units": "m",
        "standard_name": "projection_y_coordinate",
        "long_name": "y-coordinate of mesh nodes",
        "mesh": "mesh2d",
        "location": "node",
    }
    ds2["mesh2d_node_y"] = xr.DataArray(
        node_y, dims=(mesh2d_attrs["node_dimension"]), attrs=node_y_attrs
    )

    fnc = np.array(
        [
            [1, 7, 2, -999],
            [3, 2, 7, -999],
            [8, 4, 3, -999],
            [1, 6, 7, -999],
            [8, 3, 7, -999],
            [9, 4, 8, -999],
            [5, 10, 6, -999],
            [8, 7, 6, 12],
            [11, 9, 8, -999],
            [10, 12, 6, -999],
            [11, 8, 12, -999],
            [10, 13, 12, -999],
            [11, 12, 13, -999],
        ],
        dtype=int,
    )
    fnc_attrs = {
        "_FillValue": -999,
        "cf_role": "face_node_connectivity",
        "start_index": 1,
        "coordinates": "mesh2d_face_x mesh2d_face_y",
    }
    ds2["mesh2d_face_nodes"] = xr.DataArray(
        fnc,
        dims=(mesh2d_attrs["face_dimension"], mesh2d_attrs["max_face_nodes_dimension"]),
        attrs=fnc_attrs,
    )

    enc = np.array(
        [
            [1, 2],
            [1, 6],
            [1, 7],
            [2, 3],
            [2, 7],
            [3, 4],
            [3, 7],
            [3, 8],
            [4, 8],
            [4, 9],
            [5, 6],
            [5, 10],
            [6, 7],
            [6, 10],
            [6, 12],
            [7, 8],
            [8, 9],
            [8, 11],
            [8, 12],
            [9, 11],
            [10, 12],
            [10, 13],
            [11, 12],
            [11, 13],
            [12, 13],
        ],
        dtype=int,
    )
    enc_attrs = {
        "cf_role": "edge_node_connectivity",
        "mesh": "mesh2d",
        "location": "edge",
        "long_name": "Mapping from every edge to the two nodes that it connects",
        "start_index": 1,
        "_FillValue": -999,
    }
    ds2["mesh2d_edge_nodes"] = xr.DataArray(
        enc, dims=(mesh2d_attrs["edge_dimension"], "two"), attrs=enc_attrs
    )

    # add dummy face variable in order to have a face dimension in the uds
    facevar = np.ones(shape=(ds2.sizes[mesh2d_attrs["face_dimension"]]))
    ds2["mesh2d_facevar"] = xr.DataArray(facevar, dims=(mesh2d_attrs["face_dimension"]))

    # add dummy nodevar to plot and trigger triangulation procedure
    nodevar = np.ones(shape=(ds2.sizes[mesh2d_attrs["node_dimension"]]))
    ds2["mesh2d_nodevar"] = xr.DataArray(nodevar, dims=(mesh2d_attrs["node_dimension"]))
    return ds2


def get_ugrid_fillvaluem999_startindex1_uds():
    # upon loading a dataset from a file, xarray decodes it, so we also do it here
    ds2 = get_ugrid_fillvaluem999_startindex1_ds()
    ds2_enc = xr.decode_cf(ds2)
    uds = xugrid.UgridDataset(ds2_enc)
    return uds


def test_fm_fillvalue_startindex_isel():
    """
    FM data has 1-based starting index and _FillValue -999, this raises several
    issues. Since it is not possible to generate a Ugrid2d with these
    attributes, we are testing with raw data
    """

    # xugrid 0.5.0 warns "RuntimeWarning: invalid value encountered in cast: cast = data.astype(dtype, copy=True)"
    uds = get_ugrid_fillvaluem999_startindex1_uds()

    # xugrid 0.6.0 raises "ValueError: Invalid edge_node_connectivity"
    uds.isel({uds.grid.face_dimension: [1]})


def test_alternative_fill_value_start_index():
    uds = get_ugrid_fillvaluem999_startindex1_uds()
    # Check internal fill value. Should be FILL_VALUE
    grid = uds.ugrid.grid
    assert grid.face_node_connectivity.dtype == "int64"
    assert grid.start_index == 1
    assert grid.fill_value == -999
    assert (grid.face_node_connectivity != -999).all()
    gridds = grid.to_dataset()
    # Should be set back to the origina fill value.
    faces = gridds["mesh2d_face_nodes"]
    assert faces.attrs["start_index"] == 1
    uniq = np.unique(faces)
    assert uniq[0] == -999
    assert uniq[1] == 1

    # And similarly for the UgridAccessors.
    ds = uds.ugrid.to_dataset()
    faces = ds["mesh2d_face_nodes"]
    assert faces.attrs["start_index"] == 1
    uniq = np.unique(faces)
    assert uniq[0] == -999
    assert uniq[1] == 1

    ds_uda = uds["mesh2d_facevar"].ugrid.to_dataset()
    faces = ds_uda["mesh2d_face_nodes"]
    assert faces.attrs["start_index"] == 1
    uniq = np.unique(faces)
    assert uniq[0] == -999
    assert uniq[1] == 1

    # Alternative value
    grid.start_index = 0
    grid.fill_value = -2
    gridds = grid.to_dataset()
    # Should be set back to the origina fill value.
    faces = gridds["mesh2d_face_nodes"]
    assert faces.attrs["start_index"] == 0
    uniq = np.unique(faces)
    assert uniq[0] == -2
    assert uniq[1] == 0


def test_fm_facenodeconnectivity_fillvalue():
    """
    FM data has 1-based starting index and _FillValue -999, this raises several
    issues. Since it is not possible to generate a Ugrid2d with these
    attributes, we are testing with raw data
    """

    # xugrid 0.5.0 warns "RuntimeWarning: invalid value encountered in cast: cast = data.astype(dtype, copy=True)"
    uds = get_ugrid_fillvaluem999_startindex1_uds()

    # xugrid 0.6.0 has -2 values in the array
    assert (uds.grid.face_node_connectivity != -2).all()


def test_periodic_conversion():
    vertices = np.array(
        [
            [0.0, 0.0],  # 0
            [1.0, 0.0],  # 1
            [2.0, 0.0],  # 2
            [3.0, 0.0],  # 3
            [0.0, 1.0],  # 4
            [1.0, 1.0],  # 5
            [2.0, 1.0],  # 6
            [3.0, 1.0],  # 7
            [0.0, 2.0],  # 8
            [1.0, 2.0],  # 9
            [2.0, 2.0],  # 10
            [3.0, 2.0],  # 11
        ]
    )
    faces = np.array(
        [
            [0, 1, 5, 4],
            [1, 2, 6, 5],
            [2, 3, 7, 6],
            [4, 5, 9, 8],
            [5, 6, 10, 9],
            [6, 7, 11, 10],
        ]
    )
    grid = xugrid.Ugrid2d(*vertices.T, -1, faces)
    da = xr.DataArray([0, 1, 2, 3, 4, 5], dims=(grid.face_dimension,))
    uda = xugrid.UgridDataArray(da, grid)
    periodic = uda.ugrid.to_periodic()
    back = periodic.ugrid.to_nonperiodic(xmax=3.0)
    assert isinstance(periodic, xugrid.UgridDataArray)
    assert isinstance(back, xugrid.UgridDataArray)
    back_grid = back.ugrid.grid
    assert back_grid.n_face == grid.n_face
    assert back_grid.n_edge == grid.n_edge
    assert back_grid.n_node == grid.n_node

    # Also test a multi-topology dataset The 1D grid should be skipped: it
    # doesn't implement anything for these conversions, but should simply be
    # added as-is to the result.
    uds = ugrid1d_ds()
    uds["a2d"] = uda
    periodic_ds = uds.ugrid.to_periodic()
    back_ds = periodic_ds.ugrid.to_nonperiodic(xmax=3.0)
    assert isinstance(periodic_ds, xugrid.UgridDataset)
    assert isinstance(back_ds, xugrid.UgridDataset)
    assert "a1d" in back_ds
    assert "a2d" in back_ds


def test_laplace_interpolate_facets():
    grid = GRID()
    node_uda = xugrid.UgridDataArray(
        xr.DataArray(np.ones(grid.n_node), dims=(grid.node_dimension,)),
        grid=grid,
    )
    edge_uda = xugrid.UgridDataArray(
        xr.DataArray(np.ones(grid.n_edge), dims=(grid.edge_dimension,)),
        grid=grid,
    )
    face_uda = xugrid.UgridDataArray(
        xr.DataArray(np.ones(grid.n_face), dims=(grid.face_dimension,)),
        grid=grid,
    )
    node_uda[:-1] = np.nan
    edge_uda[:-1] = np.nan
    face_uda[:-1] = np.nan

    for uda in (node_uda, face_uda):
        actual = uda.ugrid.laplace_interpolate(direct_solve=True)
        assert isinstance(actual, xugrid.UgridDataArray)
        assert np.allclose(actual, 1.0)

    for uda in (node_uda, face_uda):
        actual = uda.ugrid.laplace_interpolate(direct_solve=False)
        assert isinstance(actual, xugrid.UgridDataArray)
        assert np.allclose(actual, 1.0)

    msg = "Laplace interpolation along edges is not allowed."
    with pytest.raises(ValueError, match=msg):
        edge_uda.ugrid.laplace_interpolate(direct_solve=True)

    for uda in (node_uda, edge_uda, face_uda):
        actual = uda.ugrid.interpolate_na()
        assert isinstance(actual, xugrid.UgridDataArray)
        assert np.allclose(actual, 1.0)


def test_to_facets_1d():
    uds = ugrid1d_ds()
    with pytest.raises(ValueError, match="Cannot map to face"):
        uds["a1d"].ugrid.to_face()
    with pytest.raises(ValueError, match="No conversion needed"):
        uds["a1d"].ugrid.to_node()
    with pytest.raises(ValueError, match="No conversion needed"):
        uds["b1d"].ugrid.to_edge()

    grid = uds.ugrid.grid
    to_edge = uds["a1d"].ugrid.to_edge()
    to_node = uds["b1d"].ugrid.to_node()
    assert isinstance(to_edge, xugrid.UgridDataArray)
    assert isinstance(to_node, xugrid.UgridDataArray)
    assert to_edge.dims == (grid.edge_dimension, "nmax")
    assert to_node.dims == (grid.node_dimension, "nmax")

    # Multidim
    dim_multiplier = xr.DataArray([[1, 2, 3]], dims=("dim0", "dim1"))
    uda = uds["a1d"] * dim_multiplier
    to_edge_nd = uda.ugrid.to_edge()
    assert to_edge_nd.dims == (grid.edge_dimension, "nmax", "dim0", "dim1")
    to_edge_ndT = uda.transpose().ugrid.to_edge()
    assert to_edge_ndT.dims == ("dim1", "dim0", grid.edge_dimension, "nmax")


def test_laplace_interpolate_1d():
    uda = ugrid1d_ds()["a1d"]
    uda[:] = 1.0
    uda[1] = np.nan
    actual = uda.ugrid.laplace_interpolate(direct_solve=True)
    assert isinstance(actual, xugrid.UgridDataArray)
    assert np.allclose(actual, 1.0)

    actual = uda.ugrid.laplace_interpolate(direct_solve=False)
    assert isinstance(actual, xugrid.UgridDataArray)
    assert np.allclose(actual, 1.0)


def test_laplace_interpolate_1d__disconnected():
    """
    Test laplace_interpolate on a 1D grid with disconnected edges. Should throw
    no ZeroDivisionErrors, and should preserve NaNs on the disconnected group.
    """
    xy = np.array(
        [
            [0.0, 0.0],
            [1.0, 1.0],
            [2.0, 2.0],
            [3.0, 3.0],
            [4.0, 4.0],
        ]
    )
    grid = xugrid.Ugrid1d(
        node_x=xy[:, 0],
        node_y=xy[:, 1],
        fill_value=-1,
        edge_node_connectivity=np.array([[0, 1], [1, 2], [3, 4]]),
    )
    ds = grid.to_dataset()
    ds["a1d"] = xr.DataArray(
        [1.0, np.nan, 0.0, np.nan, np.nan], dims=[grid.node_dimension]
    )
    ds["b1d"] = xr.DataArray([1.0, 2.0, 3.0], dims=[grid.edge_dimension])
    uda = xugrid.UgridDataset(ds)["a1d"]

    actual = uda.ugrid.laplace_interpolate(direct_solve=True)
    assert isinstance(actual, xugrid.UgridDataArray)
    np.testing.assert_allclose(actual[:3], np.array([1.0, 0.5, 0.0]))
    assert np.isnan(actual[3:]).all()

    actual = uda.ugrid.laplace_interpolate(direct_solve=False)
    assert isinstance(actual, xugrid.UgridDataArray)
    np.testing.assert_allclose(actual[:3], np.array([1.0, 0.5, 0.0]))
    assert np.isnan(actual[3:]).all()


def test_interpolate_na_1d():
    uda = ugrid1d_ds()["a1d"]
    with pytest.raises(ValueError, match='"abc" is not a valid interpolator.'):
        uda.ugrid.interpolate_na(method="abc")

    # Node data
    uda = ugrid1d_ds()["a1d"]
    uda[:] = 1.0
    uda[1] = np.nan
    actual = uda.ugrid.interpolate_na()
    assert isinstance(actual, xugrid.UgridDataArray)
    assert np.allclose(actual, 1.0)

    # Edge data
    uda = ugrid1d_ds()["b1d"]
    uda[:] = 1.0
    uda[1] = np.nan
    actual = uda.ugrid.interpolate_na()
    assert isinstance(actual, xugrid.UgridDataArray)
    assert np.allclose(actual, 1.0)

    # Check max_distance
    actual = uda.ugrid.interpolate_na(max_distance=0.5)
    assert np.isnan(actual[1])


def test_ugriddataset_wrap_twice(tmp_path):
    """
    In issue https://github.com/Deltares/xugrid/issues/208 wrapping a ds
    twice with UgridDataset resulted in "ValueError: connectivity contains negative values",
    because the original connectivity array in the xarray dataset was altered.
    This tests ensures that future changes will not cause this issue again.
    """
    ds_raw = get_ugrid_fillvaluem999_startindex1_ds()
    file_nc = tmp_path / "ugrid_fillvaluem999_startindex1.nc"
    ds_raw.to_netcdf(file_nc)
    ds = xr.open_dataset(file_nc)

    _ = xugrid.UgridDataset(ds)
    _ = xugrid.UgridDataset(ds)



# __init__.py for pytest-cov
import pytest


def _importorskip(modname):
    try:
        import meshkernel

        # If the DLL/SO fails to load / be found, still skip.
        try:
            meshkernel.MeshKernel()
            has = True
        except OSError:
            has = False
    except ImportError:
        has = False
    func = pytest.mark.skipif(not has, reason=f"requires {modname}")
    return has, func


has_meshkernel, requires_meshkernel = _importorskip("meshkernel")



from xugrid.data.sample_data import (
    adh_san_diego,
    elevation_nl,
    hydamo_network,
    provinces_nl,
    xoxo,
)
from xugrid.data.synthetic import disk, generate_disk

__all__ = (
    "adh_san_diego",
    "elevation_nl",
    "provinces_nl",
    "xoxo",
    "disk",
    "generate_disk",
    "hydamo_network",
)



from enum import EnumMeta, IntEnum
from typing import Union

import numpy as np


def either_string_or_enum(value: Union[str, IntEnum], enum_class: EnumMeta) -> IntEnum:
    """Convert to enum if needed, check value"""
    if isinstance(value, str):
        name = value.upper()
        enum_dict = dict(enum_class.__members__)
        try:
            value = enum_dict[name]
        except KeyError:
            valid_options = ", ".join(enum_dict.keys()).lower()
            raise ValueError(
                f"Invalid option: {value}. Valid options are: {valid_options}"
            )
    elif not isinstance(value, enum_class):
        raise TypeError(
            f"Option should be one of {enum_class}, received: {type(value)}"
        )
    return value


def to_geometry_list(
    polygon: Union["shapely.Geometry"],  # type: ignore # noqa
) -> "meshkernel.GeometryList":  # type: ignore # noqa
    import meshkernel
    import shapely

    xy = shapely.get_coordinates(shapely.get_exterior_ring(polygon))
    return meshkernel.GeometryList(np.array(xy[:, 0]), np.array(xy[:, 1]))



"""Logic for snapping points and lines to face nodes and face edges."""

from typing import Tuple, TypeVar, Union

import numba as nb
import numpy as np
import pandas as pd
import xarray as xr
from scipy.spatial import cKDTree

import xugrid as xu
from xugrid.constants import (
    FloatArray,
    IntArray,
    IntDType,
    LineArray,
    MissingOptionalModule,
    Point,
    Vector,
)
from xugrid.core.sparse import MatrixCSR, columns_and_values, row_slice
from xugrid.ugrid import connectivity
from xugrid.ugrid.connectivity import AdjacencyMatrix
from xugrid.ugrid.ugrid2d import Ugrid2d

try:
    import geopandas as gpd

    GeoDataFrameType = gpd.GeoDataFrame
except ImportError:
    gpd = MissingOptionalModule("geopandas")
    # https://stackoverflow.com/questions/61384752/how-to-type-hint-with-an-optional-import
    GeoDataFrameType = TypeVar("GeoDataFrameType")  # avoid ImportError in typehints

try:
    import shapely
except ImportError:
    shapely = MissingOptionalModule("shapely")


@nb.njit(cache=True)
def _snap_to_nearest(A: MatrixCSR, snap_candidates: IntArray, max_distance) -> IntArray:
    """
    Find a closest target for each node.

    The kD tree distance matrix will have stored for each node the other nodes
    that are within snapping distance. These are the rows in the sparse matrix
    that have more than one entry: the snap_candidates.

    The first condition for a point to become a TARGET is if it hasn't been
    connected to another point yet, i.e. it is UNVISITED. Once a point becomes
    an TARGET, it looks for nearby points within the max_distance. These nearby
    points are connected if: they are UNVISITED (i.e. they don't have a target
    yet), or the current target is closer than the previous.
    """
    UNVISITED = -1
    TARGET = -2
    nearest = np.full(A.n, max_distance + 1.0)
    visited = np.full(A.n, UNVISITED)

    for i in snap_candidates:
        if visited[i] != UNVISITED:
            continue
        visited[i] = TARGET

        # Now iterate through every node j that is within max_distance of node i.
        for j, dist in columns_and_values(A, row_slice(A, i)):
            if i == j or visited[j] == TARGET:
                # Continue if we're looking at the distance to ourselves
                # (i==j), or other node is a target.
                continue
            if visited[j] == UNVISITED or dist < nearest[j]:
                # If unvisited node, or already visited but we're closer, set
                # to i.
                visited[j] = i
                nearest[j] = dist

    return visited


def snap_nodes(
    x: FloatArray, y: FloatArray, max_snap_distance: float
) -> Tuple[FloatArray, FloatArray, IntArray]:
    """
    Snap neigbhoring vertices together that are located within a maximum
    snapping distance from each other.

    If vertices are located within a maximum distance, some of them are snapped
    to their neighbors ("targets"), thereby guaranteeing a minimum distance
    between nodes in the result. The determination of whether a point becomes a
    target itself or gets snapped to another point is primarily based on the
    order in which points are processed and their spatial relationships.

    This function also return an inverse index array. In case of a connectivity
    array, ``inverse`` can be used to index into, yielding the updated
    numbers. E.g.:

    ``updated_face_nodes = inverse[face_nodes]``

    Parameters
    ----------
    x: 1D nd array of floats of size N
    y: 1D nd array of floats of size N
    max_snap_distance: float

    Returns
    -------
    inverse: 1D nd array of ints of size N
        Inverse index array: the new vertex number for every old vertex. Is
        None when no vertices within max_distance of each other.
    x_snapped: 1D nd array of floats of size M
        Returns a copy of ``x`` when no vertices within max_distance of each
        other.
    y_snapped: 1D nd array of floats of size M
        Returns a copy of ``y`` when no vertices within max_distance of each
        other.
    """
    # First, find all the points that lie within max_distance of each other
    coords = np.column_stack((x, y))
    tree = cKDTree(coords)
    distances = tree.sparse_distance_matrix(
        tree, max_distance=max_snap_distance, output_type="coo_matrix"
    ).tocsr()
    should_snap = distances.getnnz(axis=1) > 1

    if should_snap.any():
        index = np.arange(x.size)
        visited = _snap_to_nearest(
            A=MatrixCSR.from_csr_matrix(distances),
            snap_candidates=index[should_snap],
            max_distance=max_snap_distance,
        )
        targets = visited < 0  # i.e. still UNVISITED or TARGET valued.
        visited[targets] = index[targets]
        deduplicated, inverse = np.unique(visited, return_inverse=True)
        return inverse, x[deduplicated], y[deduplicated]
    else:
        return None, x.copy(), y.copy()


def snap_to_nodes(
    x: FloatArray,
    y: FloatArray,
    to_x: FloatArray,
    to_y: FloatArray,
    max_distance: float,
    tiebreaker=None,
) -> Tuple[FloatArray, FloatArray, IntArray]:
    """
    Snap vertices (x, y) to another set of vertices (to_x, to_y) if within a
    specified maximum distance.

    Parameters
    ----------
    x: 1D nd array of floats of size N
    y: 1D nd array of floats of size N
    to_x: 1D nd array of floats of size M
    to_y: 1D nd array of floats of size M
    max_distance: float

    Returns
    -------
    x_snapped: 1D nd array of floats of size N
    y_snapped: 1D nd array of floats of size N
    """
    if tiebreaker not in (None, "nearest"):
        raise ValueError(
            f"Invalid tiebreaker: {tiebreaker}, should be one of "
            '{None, "nearest"} instead.'
        )
    # Build KDTrees
    coords = np.column_stack((x, y))
    to_coords = np.column_stack((to_x, to_y))
    tree = cKDTree(coords)
    to_tree = cKDTree(to_coords)

    # Build distance matrix, identify points to snap (update) and possible ties
    # row (i) is index in x, y; column (j) is index in to_x, to_y
    # Convert to csr for quicker row indexing, and row reductions.
    distances = tree.sparse_distance_matrix(
        to_tree, max_distance=max_distance, output_type="coo_matrix"
    ).tocsr()
    n_per_row = distances.getnnz(axis=1)
    update = n_per_row == 1
    tie = n_per_row > 1

    # Update the points that snap to a single (to_x, to_y) vertex
    xnew = x.copy()
    ynew = y.copy()
    j_update = distances[update].indices
    xnew[update] = to_x[j_update]
    ynew[update] = to_y[j_update]

    # Resolve ties with multiple (to_x, to_y) candidates
    if tie.any():
        if tiebreaker == "nearest":
            ties = distances[tie].tocoo()
            j_nearest = (
                pd.DataFrame({"i": ties.row, "distance": ties.data}, index=ties.col)
                .groupby("i")["distance"]
                .idxmin()
                .to_numpy()
            )
            xnew[tie] = to_x[j_nearest]
            ynew[tie] = to_y[j_nearest]
        elif tiebreaker is None:
            raise ValueError(
                "Ties detected: multiple options to snap to, given max distance: "
                "set a smaller tolerance or specify a tiebreaker."
            )
        else:
            raise ValueError("Invalid tiebreaker")

    return xnew, ynew


@nb.njit(inline="always")
def to_vector(a: Point, b: Point) -> Vector:
    return Vector(b.x - a.x, b.y - a.y)


@nb.njit(inline="always")
def as_point(a: FloatArray) -> Point:
    return Point(a[0], a[1])


def lines_as_edges(line_coords, line_index) -> FloatArray:
    edges = np.empty((len(line_coords) - 1, 2, 2))
    edges[:, 0, :] = line_coords[:-1]
    edges[:, 1, :] = line_coords[1:]
    keep = np.diff(line_index) == 0
    return edges[keep], line_index[1:][keep]


@nb.njit(inline="always")
def left_of(a: Point, p: Point, U: Vector) -> bool:
    # Whether point a is left of vector U
    # U: p -> q direction vector
    return U.x * (a.y - p.y) > U.y * (a.x - p.x)


def coerce_geometry(lines: GeoDataFrameType) -> LineArray:
    geometry = lines.geometry.to_numpy()
    geom_type = shapely.get_type_id(geometry)
    if not ((geom_type == 1) | (geom_type == 2)).all():
        raise ValueError("Geometry should contain only LineStrings and/or LinearRings")
    return geometry


@nb.njit(cache=True)
def snap_to_edges(
    face_indices: IntArray,
    intersection_edges: FloatArray,
    face_edge_connectivity: AdjacencyMatrix,
    edge_face_connectivity: IntArray,
    centroids: FloatArray,
    edges: IntArray,
    segment_index: IntArray,
    tolerance: float,
) -> Tuple[IntArray, IntArray]:
    """
    Snap the intersected edges to the edges of the surrounding face.

    This algorithm works as follows:

    * It takes the intersected edges; any edge (p to q) to test falls fully
      within a single face.
    * For a face, we take the centroid (a).
    * We loop through every edge of the face.
    * If the edge separates the centroid (a) from the centroid of the other
      connected face (b) we store that edge as a separating edge.

    We test for separation by:

    * Finding whether a and b are located at opposite sides of the half-plane
      created by the edge p -> q (U).
    * Finding whether p and q are located at opposide sides of the half-plane
      created by a -> b (V).
    * The separation test will return False if the lines are collinear. This is
      desirable here, if U runs collinear with V, U doesn't separate a from b.

    Do a minimum amount of work: reuse a_left, only compute V if needed.
    """
    count = 0
    for i in range(len(face_indices)):
        face = face_indices[i]
        a = as_point(centroids[face])
        p = as_point(intersection_edges[i, 0])
        q = as_point(intersection_edges[i, 1])
        U = to_vector(p, q)
        if U.x == 0 and U.y == 0:
            continue

        # Slightly enlargen the vector for edge cases.
        # Note: np.sign function returns a 0 on values of -0 and 0.
        signx = np.sign(U.x)
        signy = np.sign(U.y)
        increase = tolerance * max(abs(U.x), abs(U.y))
        p = Point(p.x - signx * increase, p.y - signy * increase)
        q = Point(q.x + signx * increase, q.y + signy * increase)
        U = to_vector(p, q)

        a_left = left_of(a, p, U)
        for edge in connectivity.neighbors(face_edge_connectivity, face):
            face_a, face_b = edge_face_connectivity[edge]
            if face_b == face:
                face_b = face_a
            if face_b == -1:
                continue

            b = as_point(centroids[face_b])
            b_left = left_of(b, p, U)
            if a_left != b_left:
                V = to_vector(a, b)
                if left_of(p, a, V) != left_of(q, a, V):
                    segment_index[count] = i
                    edges[count] = edge
                    count += 1

    return edges[:count], segment_index[:count]


def _create_output_dataset(
    lines: GeoDataFrameType,
    topology: "xu.Ugrid2d",
    edges: IntArray,
    line_index: IntArray,
) -> xu.UgridDataset:
    uds = xu.UgridDataset(grids=[topology])
    data = np.full(topology.n_edge, np.nan)
    data[edges] = line_index
    uds["line_index"] = xr.DataArray(
        data=data,
        dims=[topology.edge_dimension],
    )
    for column in lines.columns:
        if column == "geometry":
            continue
        data = np.full(topology.n_edge, np.nan)
        data[edges] = lines[column].iloc[line_index]
        uds[column] = xr.DataArray(
            data=data,
            dims=[topology.edge_dimension],
        )
    return uds


def _create_output_gdf(
    lines,
    vertices,
    edge_node_connectivity,
    edges,
    shapely_index,
):
    edge_vertices = vertices[edge_node_connectivity[edges]]
    geometry = shapely.linestrings(edge_vertices)
    return gpd.GeoDataFrame(
        lines.drop(columns="geometry").iloc[shapely_index], geometry=geometry
    )


def create_snap_to_grid_dataframe(
    lines: GeoDataFrameType,
    grid: Union[xr.DataArray, xu.UgridDataArray],
    max_snap_distance: float,
    tolerance: float = 1e-12,
) -> pd.DataFrame:
    """
    Create a dataframe required to snap line geometries to a Ugrid2d topology.

    A line is included and snapped to a grid edge when the line separates
    the centroid of the cell with the centroid of the edge.

    Parameters
    ----------
    lines: gpd.GeoDataFrame
        Line data. Geometry colum should contain exclusively LineStrings.
    grid: xugrid.Ugrid2d
        Grid of cells to snap lines to.
    max_snap_distance: float
    tolerance: float, optional, default value is 1e-12.
        Relative tolerance value to resolve edge cases. Increase the value if a
        line is unexpectedly not snapped to a grid cell edge.

    Returns
    -------
    result: pd.DataFrame
        DataFrame with columns:

        * line_index: the index of the geodataframe geometry.
        * edge_index: the index of the
        * x0: start x-coordinate of edge segment.
        * y0: start y-coordinate of edge segment.
        * x1: end x-coordinate of edge segment.
        * y1: end y-coordinate of edge segment.
        * length: length of the edge.

    Examples
    --------
    First create data frame:

    >>> snapping_df = create_snap_to_grid_dataframe(lines, grid2d, max_snap_distance=0.5)

    Use the ``line_index`` column to assign values from ``lines`` to this new dataframe:

    >>> snapping_df["my_variable"] = lines["my_variable"].iloc[snapping_df["line_index"]].to_numpy()

    Run some reduction on the variable, to create an aggregated value per grid edge:

    >>> aggregated = snapping_df.groupby("edge_index").sum()

    Assign the aggregated values to a Ugrid2d topology:

    >>> new = xu.full_like(edge_data, np.nan)
    >>> new.data[aggregated.index] = aggregated["my_variable"]

    """
    if not isinstance(grid, Ugrid2d):
        raise TypeError(f"Expected Ugrid2d, received: {type(grid).__name__}")

    topology = grid
    vertices = topology.node_coordinates
    face_edge_connectivity = topology.face_edge_connectivity
    A = connectivity.to_sparse(face_edge_connectivity)
    n, m = A.shape
    face_edge_connectivity = AdjacencyMatrix(A.indices, A.indptr, A.nnz, n, m)
    edge_face_connectivity = topology.edge_face_connectivity

    # Create geometric data
    line_geometry = coerce_geometry(lines)
    line_coords, shapely_vertex_index = shapely.get_coordinates(
        line_geometry, return_index=True
    )
    # Snap line_coords to grid
    x, y = snap_to_nodes(
        *line_coords.T, *vertices.T, max_snap_distance, tiebreaker="nearest"
    )
    line_edges, shapely_line_index = lines_as_edges(
        np.column_stack([x, y]), shapely_vertex_index
    )

    # Search for intersections. Every edge is potentially divided into smaller
    # segments: The segment_indices contain (repeated) values of the
    #
    # * line_index: for each segment, the index of the shapely geometry.
    # * face_indices: for each segment, the index of the topology face.
    # * segment_edges: for each segment, start and end xy-coordinates.
    #
    line_index, face_indices, segment_edges = topology.celltree.intersect_edges(
        line_edges
    )

    # Create edges from the intersected lines a: line can only snap to N - 1
    # edges for N edges of a face. Pre-allocate the arrays here. For some
    # reason, recent versions of numba refuse np.empty or np.zeros calls in
    # this module (!).
    # TODO: investigate...
    #
    # * edge_index: which edge of the topology (may contain duplicates)
    # * segment_index: which segment

    max_n_new_edges = len(face_indices) * topology.n_max_node_per_face - 1
    edge_index = np.empty(max_n_new_edges, dtype=IntDType)
    segment_index = np.empty(max_n_new_edges, dtype=IntDType)
    edge_index, segment_index = snap_to_edges(
        face_indices,
        segment_edges,
        face_edge_connectivity,
        edge_face_connectivity,
        topology.centroids,
        edge_index,  # out
        segment_index,  # out
        tolerance,
    )
    line_index = line_index[segment_index]
    segment_edges = segment_edges[segment_index]

    return pd.DataFrame(
        data={
            "line_index": shapely_line_index[line_index],
            "edge_index": edge_index,
            "x0": segment_edges[:, 0, 0],
            "y0": segment_edges[:, 0, 1],
            "x1": segment_edges[:, 1, 0],
            "y1": segment_edges[:, 1, 1],
            "length": ((segment_edges[:, 1] - segment_edges[:, 0]) ** 2).sum(axis=1),
        }
    )


def snap_to_grid(
    lines: GeoDataFrameType,
    grid: Union[xr.DataArray, xu.UgridDataArray],
    max_snap_distance: float,
) -> Tuple[IntArray, Union[pd.DataFrame, GeoDataFrameType]]:
    """
    Snap a collection of lines to a grid.

    A line is included and snapped to a grid edge when the line separates
    the centroid of the cell with the centroid of the edge.

    Parameters
    ----------
    lines: gpd.GeoDataFrame
        Line data. Geometry colum should contain exclusively LineStrings.
    grid: xr.DataArray or xu.UgridDataArray of integers
        Grid of cells to snap lines to.
    max_snap_distance: float

    Returns
    -------
    uds: UgridDataset
        Snapped line geometries as edges in a Ugrid2d topology. Contains a
        ``line_index`` variable identifying the original geodataframe line.
    gdf: gpd.GeoDataFrame
        Snapped line geometries.
    """
    if isinstance(grid, Ugrid2d):
        topology = grid
    elif isinstance(grid, xr.DataArray):
        # Convert structured to unstructured representation
        topology = Ugrid2d.from_structured(grid)
    elif isinstance(grid, xu.UgridDataArray):
        topology = grid.ugrid.grid
    else:
        raise TypeError(
            "Expected xarray.DataArray or xugrid.UgridDataArray, received: "
            f" {type(grid).__name__}"
        )

    result = create_snap_to_grid_dataframe(lines, topology, max_snap_distance)

    # When multiple line parts are snapped to the same edge, use the ones with
    # the greatest length inside the cell.
    max_edge_index = result.groupby("edge_index").idxmax()["length"].to_numpy()
    line_index = result["line_index"].to_numpy()[max_edge_index]
    edges = result["edge_index"].to_numpy()[max_edge_index]

    uds = _create_output_dataset(lines, topology, edges, line_index)
    gdf = _create_output_gdf(
        lines,
        topology.node_coordinates,
        topology.edge_node_connectivity,
        edges,
        line_index,
    )
    return uds, gdf



import numpy as np


def create_linear_index(arrays, dims):
    meshgrids = [a.ravel() for a in np.meshgrid(*arrays, indexing="ij")]
    return np.ravel_multi_index(meshgrids, dims)


def create_weights(arrays):
    meshgrids = np.meshgrid(*arrays, indexing="ij")
    weight = meshgrids[0]
    for dim_weight in meshgrids[1:]:
        weight *= dim_weight
    return weight.ravel()


def broadcast(
    source_shape,
    target_shape,
    source_indices,
    target_indices,
    weights,
):
    """
    Parameters
    ----------
    source_shape: tuple of int
    target_shape: tuple of int
    source_indices: tuple of numpy arrays of int
    target_indices: tuple of numpy arrays of int
    weights: tuple of numpy arrays of floats
    """
    source_index = create_linear_index(source_indices, source_shape)
    target_index = create_linear_index(target_indices, target_shape)
    weights = create_weights(weights)
    return source_index, target_index, weights


def alt_cumsum(a):
    """
    Alternative cumsum, always starts at 0 and omits the last value of the
    regular cumsum.
    """
    out = np.empty(a.size, a.dtype)
    out[0] = 0
    np.cumsum(a[:-1], out=out[1:])
    return out



"""
Wrap in advance instead of overloading __getattr__.

This allows for tab completion and documentation.
"""

from __future__ import annotations

import types
import warnings
from collections import ChainMap
from functools import wraps
from itertools import chain
from typing import List, Sequence, Union

import xarray as xr
from numpy.typing import ArrayLike
from pandas import RangeIndex

import xugrid
from xugrid.conversion import grid_from_dataset, grid_from_geodataframe
from xugrid.core.utils import unique_grids
from xugrid.ugrid.ugrid2d import Ugrid2d
from xugrid.ugrid.ugridbase import AbstractUgrid, UgridType, align

# Import entire module here for circular import of UgridDatasetAccessor and
# UgridDataArrayAccessor. Note: can only be used in functions (since that code
# is run at runtime).


def maybe_xugrid(obj, topology, old_indexes=None):
    if not isinstance(obj, (xr.DataArray, xr.Dataset)):
        return obj

    # Topology can either be a sequence of grids or a grid.
    if isinstance(topology, (list, set, tuple)):
        grids = {dim: grid for grid in topology for dim in grid.dims}
    else:
        grids = dict.fromkeys(topology.dims, topology)

    item_grids = unique_grids([grids[dim] for dim in obj.dims if dim in grids])

    if len(item_grids) == 0:
        return obj
    else:
        result, aligned = align(obj, item_grids, old_indexes)

        if isinstance(result, xr.DataArray):
            if len(aligned) > 1:
                raise RuntimeError("This shouldn't happen. Please open an issue.")
            return UgridDataArray(result, aligned[0])

        elif isinstance(result, xr.Dataset):
            return UgridDataset(result, aligned)


def maybe_xarray(arg):
    if isinstance(arg, (UgridDataArray, UgridDataset)):
        return arg.obj
    else:
        return arg


def wraps_xarray(method):
    @wraps(method)
    def wrapped(*args, **kwargs):
        self = args[0]
        args = [maybe_xarray(arg) for arg in args]
        kwargs = {k: maybe_xarray(v) for k, v in kwargs.items()}
        result = method(*args, **kwargs)

        # Sidestep staticmethods, classmethods: in that case self will not be a
        # xugrid type.
        if isinstance(self, (UgridDataArray, UgridDataset)):
            return maybe_xugrid(result, self.grids, self.obj.indexes)
        else:
            return result

    return wrapped


def wrap_accessor(accessor):
    # TODO: This will not add dynamic accessors, those most be included at
    # runtime instead?

    def wrapped(*args, **kwargs):
        args = [maybe_xarray(arg) for arg in args]
        kwargs = {k: maybe_xarray(v) for k, v in kwargs.items()}
        result = accessor(*args, **kwargs)
        return result

    return wrapped


def wrap(
    target_class_dict,
    source_class,
):
    FuncType = (types.FunctionType, types.MethodType)

    # Set every method, property from the xarray object to the UgridDataArray,
    # UgridDataset. Don't set everything, as this will break the objects.
    #
    # class Empty:
    #     pass
    #
    # keep = {
    #     "__eq__",
    #     "__ge__",
    #     "__gt__",
    #     "__le__",
    #     "__lt__",
    #     "__ne__",
    #     "__repr__",
    #     "__str__",
    # }
    #
    # remove = set(dir(Empty)) - keep

    remove = {
        # These members are shared by all objects:
        "__class__",
        "__delattr__",
        "__dict__",
        "__dir__",
        "__doc__",
        "__format__",
        "__getattribute__",
        "__hash__",
        "__init__",
        "__init_subclass__",
        "__module__",
        "__new__",
        "__reduce__",
        "__reduce_ex__",
        "__setattr__",
        "__sizeof__",
        "__subclasshook__",
        "__weakref__"
        # These are additionally included in xarray:
        "__getatrr__",
        "__slots__",
        "__annotations__",
    }

    attr_names = set(dir(source_class)) - remove
    all_attrs = {k: getattr(source_class, k) for k in attr_names}

    methods = {k: v for k, v in all_attrs.items() if isinstance(v, FuncType)}
    for name, method in methods.items():
        wrapped = wraps_xarray(method)
        setattr(wrapped, "__doc__", method.__doc__)
        target_class_dict[name] = wrapped

    properties = {k: v for k, v in all_attrs.items() if isinstance(v, property)}
    for name, prop in properties.items():
        wrapped = property(
            fget=wraps_xarray(prop.__get__),
            fset=wraps_xarray(prop.__set__),
            doc=prop.__doc__,
        )
        target_class_dict[name] = wrapped

    accessors = {k: v for k, v in all_attrs.items() if isinstance(v, type)}
    for name, accessor in accessors.items():
        wrapped = property(wrap_accessor(accessor))
        setattr(wrapped, "__doc__", accessor.__doc__)
        target_class_dict[name] = wrapped

    return


class DataArrayForwardMixin:
    wrap(
        target_class_dict=vars(),
        source_class=xr.DataArray,
    )


class DatasetForwardMixin:
    wrap(
        target_class_dict=vars(),
        source_class=xr.Dataset,
    )


def assign_ugrid_coords(obj, grids):
    grid_dims = ChainMap(*(grid.sizes for grid in grids))
    ugrid_dims = set(grid_dims.keys()).intersection(obj.dims)
    ugrid_coords = {dim: RangeIndex(0, grid_dims[dim]) for dim in ugrid_dims}
    obj = obj.assign_coords(ugrid_coords)
    return obj


class UgridDataArray(DataArrayForwardMixin):
    def __init__(self, obj: xr.DataArray, grid: UgridType):
        if not isinstance(obj, xr.DataArray):
            raise TypeError(
                f"obj must be xarray.DataArray. Received instead: {type(obj).__name__}"
            )
        if not isinstance(grid, AbstractUgrid):
            raise TypeError(
                "grid must be Ugrid1d or Ugrid2d. Received instead: "
                f"{type(grid).__name__}"
            )

        self._grid = grid
        self._obj = assign_ugrid_coords(obj, [grid])
        self._obj.set_close(obj._close)

    def __getattr__(self, attr):
        result = getattr(self.obj, attr)
        return maybe_xugrid(result, [self.grid])

    @property
    def obj(self):
        return self._obj

    @property
    def grid(self):
        return self._grid

    @property
    def grids(self) -> List[UgridType]:
        return [self._grid]

    @property
    def ugrid(self):
        """
        UGRID Accessor. This "accessor" makes operations using the UGRID
        topology available.
        """
        return xugrid.core.dataarray_accessor.UgridDataArrayAccessor(
            self.obj, self.grid
        )

    @staticmethod
    def from_structured2d(
        da: xr.DataArray,
        x: str | None = None,
        y: str | None = None,
        x_bounds: xr.DataArray = None,
        y_bounds: xr.DataArray = None,
    ) -> "UgridDataArray":
        """
        Create a UgridDataArray from a (structured) xarray DataArray.

        The spatial dimensions are flattened into a single UGRID face dimension.
        By default, this method looks for:

        1. "x" and "y" dimensions
        2. "longitude" and "latitude" dimensions
        3. "axis" attributes of "X" or "Y" on coordinates
        4. "standard_name" attributes of "longitude", "latitude",
           "projection_x_coordinate", or "projection_y_coordinate" on coordinate
           variables

        Parameters
        ----------
        da : xr.DataArray
            The structured data array to convert. The last two dimensions must be
            the y and x dimensions (in that order).
        x : str, optional
            Name of the UGRID x-coordinate, or x-dimension if bounds are provided.
            Defaults to None.
        y : str, optional
            Name of the UGRID y-coordinate, or y-dimension if bounds are provided.
            Defaults to None.
        x_bounds : xr.DataArray, optional
            Bounds for x-coordinates. Required for non-monotonic coordinates.
            Defaults to None.
        y_bounds : xr.DataArray, optional
            Bounds for y-coordinates. Required for non-monotonic coordinates.
            Defaults to None.

        Returns
        -------
        UgridDataArray
            The unstructured grid data array.

        Notes
        -----
        When using bounds, they should have one of these shapes:
        * x bounds: (M, 2) or (N, M, 4)
        * y bounds: (N, 2) or (N, M, 4)
        where N is the number of rows (along y) and M is columns (along x).
        Cells with NaN bounds coordinates are omitted.

        Examples
        --------
        Basic usage with default coordinate detection:

        >>> uda = xugrid.UgridDataArray.from_structured2d(data_array)

        Specifying explicit coordinate names:

        >>> uda = xugrid.UgridDataArray.from_structured2d(
        ...     data_array,
        ...     x="longitude",
        ...     y="latitude"
        ... )

        Using bounds for curvilinear grids:

        >>> uda = xugrid.UgridDataArray.from_structured2d(
        ...     data_array,
        ...     x="x_dim",
        ...     y="y_dim",
        ...     x_bounds=x_bounds_array,
        ...     y_bounds=y_bounds_array
        ... )
        """
        if da.ndim < 2:
            raise ValueError(
                "DataArray must have at least two spatial dimensions. "
                f"Found: {da.dims}."
            )
        if x_bounds is not None and y_bounds is not None:
            if x is None or y is None:
                raise ValueError("x and y must be provided for bounds")
            yx = (y, x)
            grid, index = Ugrid2d.from_structured_bounds(
                x_bounds=x_bounds.transpose(y, x, ...).to_numpy(),
                y_bounds=y_bounds.transpose(y, x, ...).to_numpy(),
                return_index=True,
            )
        else:
            # Possibly rely on inference of x and y dims.
            grid, yx = Ugrid2d.from_structured(da, x, y, return_dims=True)
            index = slice(None, None)

        face_da = (
            da.stack(  # noqa: PD013
                {grid.face_dimension: (yx)}, create_index=False
            )
            .isel({grid.face_dimension: index})
            .drop_vars(yx, errors="ignore")
        )
        return UgridDataArray(face_da, grid)

    @staticmethod
    def from_structured(
        da: xr.DataArray,
        x: str | None = None,
        y: str | None = None,
        x_bounds: xr.DataArray = None,
        y_bounds: xr.DataArray = None,
    ) -> "UgridDataArray":
        warnings.warn(
            "UgridDataArray.from_structured is deprecated and will be removed. "
            "Use UgridDataArray.from_structured2d instead.",
            FutureWarning,
            stacklevel=2,
        )
        return UgridDataArray.from_structured2d(da, x, y, x_bounds, y_bounds)

    @staticmethod
    def from_data(data: ArrayLike, grid: UgridType, facet: str) -> UgridDataArray:
        """
        Create a UgridDataArray from a grid and a 1D array of values.

        Parameters
        ----------
        data: array like
            Values for this array. Must be a ``numpy.ndarray`` or castable to
            it.
        grid: Ugrid1d, Ugrid2d
        facet: str
            With which facet to associate the data. Options for Ugrid1d are,
            ``"node"`` or ``"edge"``. Options for Ugrid2d are ``"node"``,
            ``"edge"``, or ``"face"``.

        Returns
        -------
        uda: UgridDataArray
        """
        return grid.create_data_array(data=data, facet=facet)


class UgridDataset(DatasetForwardMixin):
    def __init__(
        self,
        obj: xr.Dataset = None,
        grids: Union[UgridType, Sequence[UgridType]] = None,
    ):
        if obj is None and grids is None:
            raise ValueError("At least either obj or grids is required")

        if obj is None:
            original = ds = xr.Dataset()
        else:
            if not isinstance(obj, xr.Dataset):
                raise TypeError(
                    "obj must be xarray.Dataset. Received instead: "
                    f"{type(obj).__name__}"
                )
            # Remove UGRID data and associated data such as grid mapping,
            # since the CRS will be managed by the topology instead.
            connectivity_vars = [
                name
                for v in obj.ugrid_roles.connectivity.values()
                for name in v.values()
            ]
            grid_mapping_vars = [
                name
                for name in obj.ugrid_roles.grid_mapping_names.values()
                if name is not None
            ]
            # Make a copy of the variables and the attrs, ensure we don't modify the original,
            # since .drop_vars returns a shallow copy/view.
            ds = obj.drop_vars(
                obj.ugrid_roles.topology + connectivity_vars + grid_mapping_vars
            ).copy()
            for var in ds.variables.values():
                var.attrs = var.attrs.copy()
                var.attrs.pop("grid_mapping", None)
            original = obj

        if grids is None:
            topologies = obj.ugrid_roles.topology
            grids = [grid_from_dataset(obj, topology) for topology in topologies]
        else:
            # Make sure it's a new list
            if isinstance(grids, (list, tuple, set)):
                grids = list(grids)
            else:  # not iterable
                grids = [grids]
            # Now typecheck
            for grid in grids:
                if not isinstance(grid, AbstractUgrid):
                    raise TypeError(
                        "grid must be Ugrid1d or Ugrid2d. "
                        f"Received instead: {type(grid).__name__}"
                    )

        self._grids = grids
        self._obj = assign_ugrid_coords(ds, grids)
        # We've created a new object; the file handle will be associated with the original.
        # set_close makes sure that when close is called on the UgridDataset, that the
        # file will actually be closed (via the original).
        self._obj.set_close(original._close)

    @property
    def obj(self):
        return self._obj

    @property
    def grid(self) -> UgridType:
        # We need to do some checking. Don't duplicate that logic.
        return self.ugrid.grid

    @property
    def grids(self) -> List[UgridType]:
        return self._grids

    @property
    def ugrid(self):
        """
        UGRID Accessor. This "accessor" makes operations using the UGRID
        topology available.
        """
        return xugrid.core.dataset_accessor.UgridDatasetAccessor(self.obj, self.grids)

    def __getattr__(self, attr):
        result = getattr(self.obj, attr)
        return maybe_xugrid(result, self.grids)

    def __setitem__(self, key, value):
        # TODO: check with topology
        if isinstance(value, UgridDataArray):
            append = True
            # Check if the dimensions occur in self.
            # if they don't, the grid should be added.
            if self.grids is not None:
                alldims = set(chain.from_iterable([grid.dims for grid in self.grids]))
                matching_dims = set(value.grid.dims).intersection(alldims)
                if matching_dims:
                    append = False
                    # If they do match: the grids should match.
                    grids = {dim: grid for grid in self.grids for dim in grid.dims}
                    firstdim = next(iter(matching_dims))
                    grid_to_check = grids[firstdim]
                    if not grid_to_check.equals(value.grid):
                        raise ValueError(
                            "Grids share dimension names but do not are not identical. "
                            f"Matching dimensions: {matching_dims}"
                        )

            self.obj[key] = value.obj
            if append:
                self._grids.append(value.grid)
        else:
            self.obj[key] = value

    @staticmethod
    def from_geodataframe(geodataframe: "geopandas.GeoDataFrame"):  # type: ignore # noqa
        """
        Convert a geodataframe into the appropriate Ugrid topology and dataset.

        Parameters
        ----------
        geodataframe: gpd.GeoDataFrame

        Returns
        -------
        dataset: UGridDataset
        """
        grid = grid_from_geodataframe(geodataframe)
        ds = xr.Dataset.from_dataframe(geodataframe.drop("geometry", axis=1))
        return UgridDataset(ds, [grid])

    @staticmethod
    def from_structured2d(
        dataset: xr.Dataset, topology: dict | None = None
    ) -> "UgridDataset":
        """
        Create a UgridDataset from a (structured) xarray Dataset.

        The spatial dimensions are flattened into a single UGRID face dimension.
        By default, this method looks for:

        1. "x" and "y" dimensions
        2. "longitude" and "latitude" dimensions
        3. "axis" attributes of "X" or "Y" on coordinates
        4. "standard_name" attributes of "longitude", "latitude",
           "projection_x_coordinate", or "projection_y_coordinate" on coordinate
           variables

        Parameters
        ----------
        dataset : xr.Dataset
            The structured dataset to convert.
        topology : dict, optional
            Either:
            * A mapping of topology name to (x, y) coordinate names
            * A mapping of topology name to a dict containing:
            - "x": x-dimension name
            - "y": y-dimension name
            - "bounds_x": x-bounds variable name
            - "bounds_y": y-bounds variable name
            Defaults to {"mesh2d": (None, None)}.

        Returns
        -------
        UgridDataset
            The unstructured grid dataset.

        Notes
        -----
        When using bounds, they should have one of these shapes:
        * x bounds: (M, 2) or (N, M, 4)
        * y bounds: (N, 2) or (N, M, 4)
        where N is the number of rows (along y) and M is columns (along x).
        Cells with NaN bounds coordinates are omitted.

        Examples
        --------
        Basic usage with default coordinate names:

        >>> uds = xugrid.UgridDataset.from_structured2d(dataset)

        Specifying custom coordinate names:

        >>> uds = xugrid.UgridDataset.from_structured2d(
        ...     dataset,
        ...     topology={"my_mesh2d": {"x": "xc", "y": "yc"}}
        ... )

        Multiple grid topologies in a single dataset:

        >>> uds = xugrid.UgridDataset.from_structured2d(
        ...     dataset,
        ...     topology={
        ...         "mesh2d_xy": {"x": "x", "y": "y"},
        ...         "mesh2d_lonlat": {"x": "lon", "y": "lat"}
        ...     }
        ... )

        Using bounds for non-monotonic coordinates (e.g., curvilinear grids):

        >>> uds = xugrid.UgridDataset.from_structured2d(
        ...     dataset,
        ...     topology={
        ...         "my_mesh2d": {
        ...             "x": "M",
        ...             "y": "N",
        ...             "bounds_x": "grid_x",
        ...             "bounds_y": "grid_y"
        ...         }
        ...     }
        ... )
        """
        if topology is None:
            # By default, set None. This communicates to
            # Ugrid2d.from_structured to infer x and y dims.
            topology = {"mesh2d": (None, None)}

        grids = []
        dss = []
        xy_vars = set()  # store x, y, x_bounds, y_bounds to drop.
        for name, args in topology.items():
            x_bounds = None
            y_bounds = None
            if isinstance(args, dict):
                x = args.get("x")
                y = args.get("y")
                if "x_bounds" in args and "y_bounds" in args:
                    if x is None or y is None:
                        raise ValueError("x and y must be provided for bounds")
                    x_bounds = dataset[args["x_bounds"]]
                    y_bounds = dataset[args["y_bounds"]]
                    xy_vars.update((args["x_bounds"], args["y_bounds"]))
            elif isinstance(args, tuple):
                x, y = args
            else:
                raise TypeError(
                    "Expected dict or tuple in topology, received: "
                    f"{type(args).__name__}"
                )

            if x_bounds is not None and y_bounds is not None:
                stackdims = (y, x)
                grid, index = Ugrid2d.from_structured_bounds(
                    x_bounds.transpose(*stackdims, ...).to_numpy(),
                    y_bounds.transpose(*stackdims, ...).to_numpy(),
                    name=name,
                    return_index=True,
                )
            else:
                grid, stackdims = Ugrid2d.from_structured(
                    dataset, x=x, y=y, name=name, return_dims=True
                )
                index = slice(None, None)

            # Use subset to check that ALL dims of stackdims are present in the
            # variable.
            checkdims = set(stackdims)
            xy_vars.update(checkdims)
            ugrid_vars = [
                name
                for name, var in dataset.data_vars.items()
                if checkdims.issubset(var.dims) and name not in xy_vars
            ]
            dss.append(
                dataset[ugrid_vars]  # noqa: PD013
                .stack({grid.face_dimension: stackdims})
                .isel({grid.face_dimension: index})
                .drop_vars(stackdims + (grid.face_dimension,))
            )
            grids.append(grid)

        # Add the original dataset to include all non-UGRID variables.
        dss.append(dataset.drop_vars(xy_vars, errors="ignore"))
        # Then merge with compat="override". This'll pick the first available
        # variable: i.e. it will prioritize the UGRID form.
        merged = xr.merge(dss, compat="override")
        return UgridDataset(merged, grids)

    @staticmethod
    def from_structured(
        dataset: xr.Dataset, topology: dict | None = None
    ) -> "UgridDataset":
        warnings.warn(
            "UgridDataset.from_structured is deprecated and will be removed. "
            "Use UgridDataset.from_structured2d instead.",
            FutureWarning,
            stacklevel=2,
        )
        return UgridDataset.from_structured2d(dataset, topology)



import numpy as np
import xarray as xr

import xugrid


def transform(vertices, minx, maxx, miny):
    """
    Transform vertices to fit within minx to maxx.

    Maintains x:y aspect ratio.
    """
    x, y = vertices.T
    xmin = x.min()
    xmax = x.max()
    ymin = y.min()
    ymax = y.max()
    dx = xmax - xmin
    dy = ymax - ymin
    new_dx = maxx - minx
    new_dy = dy / dx * new_dx
    x = (x - xmin) * new_dx / dx + minx
    y = (y - ymin) * new_dy / dy + miny
    return np.column_stack([x, y])


def generate_disk(partitions: int, depth: int):
    """
    Generate a triangular mesh for the unit circle.

    Parameters
    ----------
    partitions: int
        Number of triangles around the origin.
    depth: int
        Number of "layers" of triangles around the origin.

    Returns
    -------
    vertices: np.ndarray of floats with shape ``(n_vertex, 2)``
    triangles: np.ndarray of integers with shape ``(n_triangle, 3)``
    """
    import matplotlib.tri

    if partitions < 3:
        raise ValueError("partitions should be >= 3")

    N = depth + 1
    n_per_level = partitions * np.arange(N)
    n_per_level[0] = 1

    delta_angle = (2 * np.pi) / np.repeat(n_per_level, n_per_level)
    index = np.repeat(np.insert(n_per_level.cumsum()[:-1], 0, 0), n_per_level)
    angles = delta_angle.cumsum()
    angles = angles - angles[index] + 0.5 * np.pi
    radii = np.repeat(np.linspace(0.0, 1.0, N), n_per_level)

    x = np.cos(angles) * radii
    y = np.sin(angles) * radii
    triang = matplotlib.tri.Triangulation(x, y)
    return np.column_stack((x, y)), triang.triangles


def disk():
    def function_z(x, y):
        """
        Generate a somewhat interesting surface.

        See: https://matplotlib.org/stable/gallery/images_contours_and_fields/tricontour_smooth_user.html
        """
        r1 = np.sqrt((0.5 - x) ** 2 + (0.5 - y) ** 2)
        theta1 = np.arctan2(0.5 - x, 0.5 - y)
        r2 = np.sqrt((-x - 0.2) ** 2 + (-y - 0.2) ** 2)
        theta2 = np.arctan2(-x - 0.2, -y - 0.2)
        z = -(
            2 * (np.exp((r1 / 10) ** 2) - 1) * 30.0 * np.cos(7.0 * theta1)
            + (np.exp((r2 / 10) ** 2) - 1) * 30.0 * np.cos(11.0 * theta2)
            + 0.7 * (x**2 + y**2)
        )
        zmin = z.min()
        zmax = z.max()
        return (zmax - z) / (zmax - zmin) * 10.0

    vertices, triangles = generate_disk(6, 8)
    vertices = transform(vertices, 0.0, 10.0, 0.0)
    grid = xugrid.Ugrid2d(
        node_x=vertices[:, 0],
        node_y=vertices[:, 1],
        fill_value=-1,
        face_node_connectivity=triangles,
    )

    ds = xr.Dataset()
    ds["node_z"] = xr.DataArray(
        data=function_z(*grid.node_coordinates.T),
        dims=[grid.node_dimension],
    )
    ds["face_z"] = xr.DataArray(
        data=function_z(*grid.face_coordinates.T),
        dims=[grid.face_dimension],
    )
    ds["edge_z"] = xr.DataArray(
        data=function_z(*grid.edge_coordinates.T),
        dims=[grid.edge_dimension],
    )
    return xugrid.UgridDataset(ds, grid)



# Skip Tcl failures
import matplotlib

matplotlib.use("Agg", force=True)

from tests.fixtures.fixture_regridder import (
    disk,
    disk_layered,
    expected_results_centroid,
    expected_results_linear,
    expected_results_overlap,
    grid_data_a,
    grid_data_a_1d,
    grid_data_a_2d,
    grid_data_a_layered,
    grid_data_a_layered_1d,
    grid_data_a_layered_2d,
    grid_data_b,
    grid_data_b_1d,
    grid_data_b_2d,
    grid_data_b_flipped_1d,
    grid_data_c,
    grid_data_c_1d,
    grid_data_c_2d,
    grid_data_d,
    grid_data_d_1d,
    grid_data_dask_expected,
    grid_data_dask_expected_layered,
    grid_data_dask_source,
    grid_data_dask_source_layered,
    grid_data_dask_target,
    grid_data_e,
    grid_data_e_1d,
    quads_0_25,
    quads_1,
    quads_structured,
)

__all__ = (
    "disk",
    "disk_layered",
    "expected_results_centroid",
    "expected_results_linear",
    "expected_results_overlap",
    "grid_data_a",
    "grid_data_a_1d",
    "grid_data_a_2d",
    "grid_data_a_layered",
    "grid_data_a_layered_1d",
    "grid_data_a_layered_2d",
    "grid_data_b",
    "grid_data_b_1d",
    "grid_data_b_2d",
    "grid_data_b_flipped_1d",
    "grid_data_c",
    "grid_data_c_1d",
    "grid_data_c_2d",
    "grid_data_d",
    "grid_data_d_1d",
    "grid_data_dask_expected",
    "grid_data_dask_expected_layered",
    "grid_data_dask_source",
    "grid_data_dask_source_layered",
    "grid_data_dask_target",
    "grid_data_e",
    "grid_data_e_1d",
    "quads_0_25",
    "quads_1",
    "quads_structured",
)



import numpy as np
import pytest
from scipy import sparse

from xugrid.ugrid import interpolate


def test_ilu0():
    # Create a 1D Laplace problem:
    #
    # * Dirichlet boundary left and right of 1.0.
    # * Constant inflow 0.001 everywhere else.
    #
    n = 1000
    d = np.ones(n)
    A = sparse.diags((-d[:-1], 2 * d, -d[:-1]), (-1, 0, 1)).tocsr()
    A[0, 0] = 1.0
    A[0, 1] = 0.0
    A[-1, -1] = 1.0
    A[-1, -2] = 0.0
    b = np.full(n, 0.001)
    b[0] = 1.0
    b[-1] = 1.0
    M = interpolate.ILU0Preconditioner.from_csr_matrix(A)
    _, info_cg = sparse.linalg.cg(A, b, maxiter=10)
    x_pcg, info_pcg = sparse.linalg.cg(A, b, maxiter=10, M=M)
    x_direct = sparse.linalg.spsolve(A, b)
    assert info_cg != 0  # cg does not converge
    assert info_pcg == 0  # preconditioned cg does converge
    assert np.allclose(x_pcg, x_direct)  # answer matches direct solve


def test_laplace_interpolate():
    i = np.array([0, 1, 1, 2, 2, 3, 3])
    j = np.array([1, 0, 2, 1, 3, 2, 4])
    coo_content = (j, (i, j))
    data = np.array([1.0, np.nan, np.nan, np.nan, 5.0])
    labels = np.zeros_like(data, dtype=int)
    with pytest.raises(ValueError, match="connectivity is not a square matrix"):
        con = sparse.coo_matrix(coo_content, shape=(4, 5)).tocsr()
        interpolate.laplace_interpolate(data, con, labels, use_weights=False)

    i = np.array([0, 1, 1, 2, 2, 3, 3, 4])
    j = np.array([1, 0, 2, 1, 3, 2, 4, 3])
    coo_content = (j, (i, j))
    data = np.array([1.0, np.nan, np.nan, np.nan, 5.0])
    expected = np.arange(1.0, 6.0)
    con = sparse.coo_matrix(coo_content, shape=(5, 5)).tocsr()
    actual = interpolate.laplace_interpolate(
        data, con, labels, use_weights=False, direct_solve=True
    )
    assert np.allclose(actual, expected)

    actual = interpolate.laplace_interpolate(
        data, con, labels, use_weights=False, direct_solve=False
    )
    assert np.allclose(actual, expected)
